from enum import Enum
from queue import Queue
from threading import Thread
from typing import Callable, Optional, List

from .errors import AsyncTaskException

class TaskStatus(str, Enum):
    PENDING = 'pending'
    SUCCESS = 'success'
    FAILURE = 'failure'

class WaitType(str, Enum):
    ALL = 'all'
    FIRST = 'first'

class ErrorHandling(str, Enum):
    RAISE = 'raise'
    IGNORE = 'ignore'

class AsyncTask:

    def __init__(
        self, 
        func: Callable, 
        *args, 
        **kwargs
    ):
        self.func: Callable = func
        self.args = args
        self.kwargs = kwargs
        self.status = TaskStatus.PENDING # 'pending'
        self.result = None
        self.thread: Thread = None
        self.exception: Exception = None
        self.on_success = None
        self.on_error = None
        self.execute_on_caller = False

    def func_handler(self):
        try:
            result = self.func(*self.args, **self.kwargs)
            self.status = TaskStatus.SUCCESS # 'success'
            self.result = result
            if self.on_success is not None and not self.execute_on_caller:
                self.on_success(self.result)
        except Exception as e:
            self.exception = e
            self.status = TaskStatus.FAILURE # 'failure'
            if self.on_error is not None and not self.execute_on_caller:
                self.on_error(self.exception)

        global_task_manager.remove(self.thread)

    def run(self):
        child_thread = Thread(target=self.func_handler)
        self.thread = child_thread
        global_task_manager.put(child_thread)

    def wait(self):
        while not self.thread.is_alive():
            pass
        self.thread.join()
        if self.exception is not None:
            raise AsyncTaskException(str(self.exception), self.func.__name__)

        return self.result

    def subscribe(self, on_success: Callable, on_error: Callable, blocking: bool = False):
        if on_success is None or on_error is None:
            raise ValueError("Illegal argument. Callbacks on_success and on_error must not be null")

        self.on_success = on_success
        self.on_error = on_error
        self.execute_on_caller = blocking
        if blocking:
            while not self.thread.is_alive(): pass
            self.thread.join()
            if self.exception is None:
                on_success(self.result)
            else:
                on_error(self.exception)


class GlobalTaskManager:
    '''
    Singleton class designed to manage async tasks in a thread pool.
    Multiple instances of this class can cause unintended behaviour
    '''

    max_threads: Optional[int] = 32

    def __init__(self, max_threads: Optional[int] = None):
        if max_threads is not None:
            # set to 200 if negative or 0
            self.max_threads = 200 if max_threads <= 0 else max_threads
        
        self.thread_pool: List[Thread] = []
        self.task_queue = Queue()

    def put(self, thread: Thread):
        self.task_queue.put(thread)

    def remove(self, thread: Thread):
        for t in self.thread_pool:
            if thread.ident == t.ident:
                self.thread_pool.remove(t)

    def run(self):
        while True:
            if len(self.thread_pool) < self.max_threads:
                child_thread: Thread = self.task_queue.get()
                child_thread.start()
                self.thread_pool.append(child_thread)


global_task_manager = GlobalTaskManager()
worker = Thread(target = global_task_manager.run, daemon = True)
worker.start()


from kubernetes.client.models import *

# from kubernetes.client.models import (
#     V1Container,
#     V1ContainerPort,
#     V1EnvVar,
#     V1EnvVarSource,

#     V1Namespace,
#     V1NamespaceList,

#     V1ConfigMap,
#     V1ConfigMapList,
#     V1ConfigMapKeySelector,

#     V1Secret,
#     V1SecretList,
#     V1SecretKeySelector,

#     V1ObjectMeta,
#     V1ObjectReference,
#     V1Pod,
#     V1PodList,
#     V1PodSpec,
#     V1PodStatus,
#     V1PodCondition,
#     V1PodTemplateSpec,
#     V1ResourceRequirements,
#     V1Service,
#     V1ServiceList,
#     V1ServicePort,
#     V1ServiceSpec,
#     V1Volume,
#     V1VolumeMount,
#     V1PersistentVolumeClaim,
#     V1PersistentVolumeClaimList,
#     V1PersistentVolumeClaimSpec,
#     V1PersistentVolumeClaimVolumeSource,
#     V1PersistentVolume,
#     V1PersistentVolumeList,
#     V1PersistentVolumeSpec,
#     V1PersistentVolumeClaimStatus,
#     V1PersistentVolumeStatus,
#     V1PersistentVolumeClaimCondition,

#     V1ContainerState,
#     V1ContainerStateRunning,
#     V1ContainerStateTerminated,
#     V1ContainerStateWaiting,
#     V1ContainerStatus,

#     V1Ingress,
#     V1IngressList,
#     V1IngressSpec,
#     V1IngressRule,
#     V1IngressTLS,
#     V1HTTPIngressPath,
#     V1HTTPIngressRuleValue,
    
#     V1Node,
#     V1NodeList,
#     V1NodeSpec,
#     V1NodeStatus,
#     V1NodeAddress,
#     V1NodeCondition,
#     V1NodeSystemInfo,
#     V1NodeDaemonEndpoints,
#     V1NodeConfigSource,
#     V1NodeConfigStatus,
#     V1NodeSelector,
#     V1NodeSelectorTerm,
#     V1NodeSelectorRequirement,
#     V1NodeAffinity,
#     V1PreferredSchedulingTerm,

#     V1Taint,
#     V1EventSource,
#     V1beta1Event,
#     V1beta1EventList,

#     V1ServiceAccount,
#     V1ServiceAccountList,

#     V1StatefulSet,
#     V1StatefulSetList,
#     V1StatefulSetSpec,
#     V1StatefulSetStatus,
# )





from __future__ import annotations

from lazyops.libs.dbinit.base import Engine

from lazyops.libs.dbinit.entities.entity import Entity
from lazyops.libs.dbinit.entities.role import Role


class Controller:
    """
    Entrypoint for creating/dropping database entities and granting/revoking privileges once they are
    declared in code. This is a wrapper of sorts around :class:`lazyops.libs.dbinit.entities.entity.Entity`.
    """

    @classmethod
    def create_all(cls, engine: Engine | None = None) -> None:
        """
        Attempts to create all declared entities. Typically run via `run_all`.
        :param engine: A :class:`sqlalchemy.Engine` that defines the connection to a Postgres instance/cluster.
        """
        cls._handle_engine(engine)
        for entity in Entity.entities:
            entity._safe_create()

    @classmethod
    def grant_all(cls, engine: Engine | None = None) -> None:
        """
        Attempts to grant all declared privileges. Requires entities to exist, typically run via `run_all` or after
        `create_all`.
        :param engine: A :class:`sqlalchemy.Engine` that defines the connection to a Postgres instance/cluster.
        """
        cls._handle_engine(engine)
        for entity in Entity.entities:
            if isinstance(entity, Role):
                entity._safe_grant()

    @classmethod
    def run_all(cls, engine: Engine | None = None) -> None:
        """
        Attempts to create all declared entities then grant all declared privileges. Main way to do so.
        :param engine: A :class:`sqlalchemy.Engine` that defines the connection to a Postgres instance/cluster.
        """
        cls._handle_engine(engine)
        cls.create_all()
        cls.grant_all()

    @classmethod
    def drop_all(cls, engine: Engine | None = None) -> None:
        """
        Attempts to drop all declared entities. Typically run via `remove_all`.
        :param engine: A :class:`sqlalchemy.Engine` that defines the connection to a Postgres instance/cluster.
        """
        cls._handle_engine(engine)
        for entity in reversed(Entity.entities):
            entity._safe_drop()

    @classmethod
    def revoke_all(cls, engine: Engine | None = None) -> None:
        """
        Attempts to revoke all declared privileges. Requires entities to exist, typically run via `remove_all` or before
        `drop_all`.
        :param engine: A :class:`sqlalchemy.Engine` that defines the connection to a Postgres instance/cluster.
        """
        cls._handle_engine(engine)
        for entity in reversed(Entity.entities):
            if isinstance(entity, Role):
                entity._safe_revoke()

    @classmethod
    def remove_all(cls, engine: Engine | None = None) -> None:
        """
        Attempts to revoke all declared privileges then drop all declared entities. Main way to do so.
        :param engine: A :class:`sqlalchemy.Engine` that defines the connection to a Postgres instance/cluster.
        """
        cls._handle_engine(engine)
        cls.revoke_all()
        cls.drop_all()

    @classmethod
    def _all_entities_exist(cls, engine: Engine | None = None) -> bool:
        """
        Checks if all declared entities exist in the cluster.
        :param engine:  A :class:`sqlalchemy.Engine` that defines the connection to a Postgres instance/cluster.
        """
        cls._handle_engine(engine)
        return all(entity._exists() for entity in Entity.entities)

    @classmethod
    def _all_grants_exist(cls, engine: Engine | None = None) -> bool:
        """
        Checks if all declared grants exist in the cluster.
        :param engine:  A :class:`sqlalchemy.Engine` that defines the connection to a Postgres instance/cluster.
        """
        cls._handle_engine(engine)
        return all(
            role._grants_exist()
            for role in Entity.entities
            if isinstance(role, Role)
        )

    @classmethod
    def _all_exist(cls, engine: Engine | None = None) -> bool:
        """
        Checks if all declared entities and grants exist in the cluster.
        :param engine:  A :class:`sqlalchemy.Engine` that defines the connection to a Postgres instance/cluster.
        """
        cls._handle_engine(engine)
        return cls._all_entities_exist() and cls._all_grants_exist()

    @staticmethod
    def _handle_engine(engine: Engine | None = None) -> None:
        """
        Utility to assign the engine if it is provided.
        :param engine: A :class:`sqlalchemy.Engine` that defines the connection to a Postgres instance/cluster.
        """
        if engine:
            Entity._engine = engine





import os
import io
import zlib
import gzip
import codecs
import struct
import sqlite3
import errno
import pickletools
import dill as pkl
import os.path as op
import functools as ft
import contextlib as cl

from fileio.lib.types import File, FileLike
from typing import Any, Iterable, AsyncIterable
from lazyops.utils.pooler import ThreadPooler
from . import base
from lazyops.libs.sqlcache.constants import (
    UNKNOWN,
    MODE_NONE,
    MODE_RAW,
    MODE_TEXT,
    MODE_BINARY,
    MODE_PICKLE,
    KeyT,
    ValueT,
    WindowsExceptionError,
)


class Disk:
    "Store key and value serialization for SQLite database and files."

    def __init__(
        self, 
        directory: FileLike = None, 
        min_file_size: int = 0, 
        pickle_protocol: int = 3, 
        **config
    ):
        """Initialize disk instance.
        :param str directory: directory path
        :param int min_file_size: minimum size for file use
        :param int pickle_protocol: pickle protocol for serialization
        :param dict config: dict to pass to update sql settings
        """
        self._directory = File(directory)
        self.min_file_size = min_file_size
        self.pickle_protocol = pickle_protocol

    def hash(self, key: KeyT):
        """Compute portable hash for `key`.
        :param key: key to hash
        :return: hash value
        """
        mask = 0xFFFFFFFF
        disk_key, _ = self.put(key)
        type_disk_key = type(disk_key)

        if type_disk_key is sqlite3.Binary:
            return zlib.adler32(disk_key) & mask
        elif type_disk_key is str:
            return zlib.adler32(disk_key.encode('utf-8')) & mask  # noqa
        elif type_disk_key is int:
            return disk_key % mask
        else:
            assert type_disk_key is float
            return zlib.adler32(struct.pack('!d', disk_key)) & mask
    
    async def ahash(self, key: KeyT):
        """Compute portable hash for `key`.
        :param key: key to hash
        :return: hash value
        """
        return await ThreadPooler.run_async(
            self.hash,
            key
        )

    def put(self, key: KeyT):
        """Convert `key` to fields key and raw for Store table.
        :param key: key to convert
        :return: (database key, raw boolean) pair
        """
        # pylint: disable=unidiomatic-typecheck
        type_key = type(key)
        if type_key is bytes:
            return sqlite3.Binary(key), True
        elif (
            (type_key is str)
            or (
                type_key is int
                and -9223372036854775808 <= key <= 9223372036854775807
            )
            or (type_key is float)
        ):
            return key, True
        else:
            data = pkl.dumps(key, protocol=self.pickle_protocol)
            result = pickletools.optimize(data)
            return sqlite3.Binary(result), False
    
    async def aput(self, key: KeyT):
        """Convert `key` to fields key and raw for Store table.
        :param key: key to convert
        :return: (database key, raw boolean) pair
        """
        return await ThreadPooler.run_async(
            self.put,
            key
        )

    def get(self, key: KeyT, raw: bool):
        """Convert fields `key` and `raw` from Store table to key.
        :param key: database key to convert
        :param bool raw: flag indicating raw database storage
        :return: corresponding Python key
        """
        # pylint: disable=no-self-use,unidiomatic-typecheck
        if raw:
            return bytes(key) if type(key) is sqlite3.Binary else key
        else:
            return pkl.load(io.BytesIO(key))
    
    async def aget(self, key: KeyT, raw: bool):
        """Convert fields `key` and `raw` from Store table to key.
        :param key: database key to convert
        :param bool raw: flag indicating raw database storage
        :return: corresponding Python key
        """
        return await ThreadPooler.run_async(
            self.get,
            key,
            raw
        )

    def store(self, value: ValueT, read: bool, key: KeyT = UNKNOWN):
        """Convert `value` to fields size, mode, filename, and value for Store
        table.
        :param value: value to convert
        :param bool read: True when value is file-like object
        :param key: key for item (default UNKNOWN)
        :return: (size, mode, filename, value) tuple for Store table
        """
        # pylint: disable=unidiomatic-typecheck
        type_value = type(value)
        min_file_size = self.min_file_size

        if (
            (type_value is str and len(value) < min_file_size)
            or (
                type_value is int
                and -9223372036854775808 <= value <= 9223372036854775807
            )
            or (type_value is float)
        ):
            return 0, MODE_RAW, None, value
        elif type_value is bytes:
            if len(value) < min_file_size:
                return 0, MODE_RAW, None, sqlite3.Binary(value)
            filename, full_path = self.filename(key, value)
            self._write(full_path, io.BytesIO(value), 'xb')
            return len(value), MODE_BINARY, filename, None
        elif type_value is str:
            filename, full_path = self.filename(key, value)
            self._write(full_path, io.StringIO(value), 'x', 'UTF-8')
            # size = op.getsize(full_path)
            size = (full_path.info())['size']
            return size, MODE_TEXT, filename, None
        elif read:
            reader = ft.partial(value.read, 2**22)
            filename, full_path = self.filename(key, value)
            iterator = iter(reader, b'')
            size = self._write(full_path, iterator, 'xb')
            return size, MODE_BINARY, filename, None
        else:
            result = pkl.dumps(value, protocol=self.pickle_protocol)

            if len(result) < min_file_size:
                return 0, MODE_PICKLE, None, sqlite3.Binary(result)
            filename, full_path = self.filename(key, value)
            self._write(full_path, io.BytesIO(result), 'xb')
            return len(result), MODE_PICKLE, filename, None
    
    async def astore(self, value: ValueT, read: bool, key: KeyT = UNKNOWN):
        """Convert `value` to fields size, mode, filename, and value for Store
        table.
        :param value: value to convert
        :param bool read: True when value is file-like object
        :param key: key for item (default UNKNOWN)
        :return: (size, mode, filename, value) tuple for Store table
        """
        # pylint: disable=unidiomatic-typecheck
        type_value = type(value)
        min_file_size = self.min_file_size

        if (
            (type_value is str and len(value) < min_file_size)
            or (
                type_value is int
                and -9223372036854775808 <= value <= 9223372036854775807
            )
            or (type_value is float)
        ):
            return 0, MODE_RAW, None, value
        elif type_value is bytes:
            if len(value) < min_file_size:
                return 0, MODE_RAW, None, sqlite3.Binary(value)
            filename, full_path = self.filename(key, value)
            await self._awrite(full_path, io.BytesIO(value), 'xb')
            return len(value), MODE_BINARY, filename, None
        elif type_value is str:
            filename, full_path = self.filename(key, value)
            await self._awrite(full_path, io.StringIO(value), 'x', 'UTF-8')
            size = (await full_path.async_info())['size']
            # size = op.getsize(full_path)
            return size, MODE_TEXT, filename, None
        elif read:
            reader = ft.partial(value.read, 2**22)
            filename, full_path = self.filename(key, value)
            iterator = iter(reader, b'')
            size = await self._awrite(full_path, iterator, 'xb')
            return size, MODE_BINARY, filename, None
        else:
            result = pkl.dumps(value, protocol=self.pickle_protocol)
            if len(result) < min_file_size:
                return 0, MODE_PICKLE, None, sqlite3.Binary(result)
            filename, full_path = self.filename(key, value)
            await self._awrite(full_path, io.BytesIO(result), 'xb')
            return len(result), MODE_PICKLE, filename, None
    
    def _write(self, full_path: FileLike, iterator: Iterable, mode: str, encoding: str = None):
        # full_dir, _ = op.split(full_path)
        full_dir = full_path.parent
        for count in range(1, 11):
            with cl.suppress(OSError):
                full_dir.mkdir(parents=True, exist_ok=True)
                # os.makedirs(full_dir)

            try:
                # Another cache may have deleted the directory before
                # the file could be opened.
                writer = full_path.open(mode = mode, encoding = encoding)
                # writer = open(full_path, mode, encoding=encoding)
            except OSError:
                if count == 10:
                    # Give up after 10 tries to open the file.
                    raise
                continue

            with writer:
                size = 0
                for chunk in iterator:
                    size += len(chunk)
                    writer.write(chunk)
                return size
    
    async def _awrite(self, full_path: FileLike, iterator: Iterable, mode: str, encoding: str = None):
        full_dir = full_path.parent
        for count in range(1, 11):
            with cl.suppress(OSError):
                full_dir.mkdir(parents=True, exist_ok=True)

            try:
                # Another cache may have deleted the directory before
                # the file could be opened.
                writer = full_path.async_open(mode = mode, encoding = encoding)
            except OSError:
                if count == 10:
                    # Give up after 10 tries to open the file.
                    raise
                continue

            async with writer:
                size = 0
                for chunk in iterator:
                    size += len(chunk)
                    await writer.write(chunk)
                return size

    def fetch(self, mode: int, filename: str, value: ValueT, read: bool):
        """Convert fields `mode`, `filename`, and `value` from Store table to
        value.
        :param int mode: value mode raw, binary, text, or pickle
        :param str filename: filename of corresponding value
        :param value: database value
        :param bool read: when True, return an open file handle
        :return: corresponding Python value
        """
        # pylint: disable=no-self-use,unidiomatic-typecheck
        if mode == MODE_RAW:
            return bytes(value) if type(value) is sqlite3.Binary else value
        elif mode == MODE_BINARY:
            if read:
                return self._directory.joinpath(filename).open('rb')
                # return open(op.join(self._directory, filename), 'rb')
            return self._directory.joinpath(filename).read_bytes()
            # with open(op.join(self._directory, filename), 'rb') as reader:
            #     return reader.read()
        elif mode == MODE_TEXT:
            return self._directory.joinpath(filename).read_text(encoding='UTF-8')
            # full_path = op.join(self._directory, filename)
            # with open(full_path, 'r', encoding='UTF-8') as reader:
            #     return reader.read()
        elif mode == MODE_PICKLE:
            if value is not None: return pkl.load(io.BytesIO(value))
            return pkl.load(self._directory.joinpath(filename).read_bytes())
            # with open(op.join(self._directory, filename), 'rb') as reader:
            #     return pkl.load(reader)

    async def afetch(self, mode: int, filename: str, value: ValueT, read: bool):
        """Convert fields `mode`, `filename`, and `value` from Store table to
        value.
        :param int mode: value mode raw, binary, text, or pickle
        :param str filename: filename of corresponding value
        :param value: database value
        :param bool read: when True, return an open file handle
        :return: corresponding Python value
        """
        if mode == MODE_RAW:
            return bytes(value) if type(value) is sqlite3.Binary else value
        elif mode == MODE_BINARY:
            if read:
                return self._directory.joinpath(filename).async_open('rb')
                # return open(op.join(self._directory, filename), 'rb')
            return await self._directory.joinpath(filename).async_read_bytes()
        
        elif mode == MODE_TEXT:
            return await self._directory.joinpath(filename).async_read_text(encoding='UTF-8')

        elif mode == MODE_PICKLE:
            if value is not None: return pkl.load(io.BytesIO(value))
            return pkl.load(await self._directory.joinpath(filename).async_read_bytes())

    def filename(self, key: KeyT = UNKNOWN, value: ValueT = UNKNOWN):
        """Return filename and full-path tuple for file storage.
        Filename will be a randomly generated 28 character hexadecimal string
        with ".val" suffixed. Two levels of sub-directories will be used to
        reduce the size of directories. On older filesystems, lookups in
        directories with many files may be slow.
        The default implementation ignores the `key` and `value` parameters.
        In some scenarios, for example :meth:`Store.push
        <diskcache.Store.push>`, the `key` or `value` may not be known when the
        item is stored in the cache.
        :param key: key for item (default UNKNOWN)
        :param value: value for item (default UNKNOWN)
        """
        # pylint: disable=unused-argument
        hex_name = codecs.encode(os.urandom(16), 'hex').decode('utf-8')
        sub_dir = op.join(hex_name[:2], hex_name[2:4])
        name = f'{hex_name[4:]}.val'
        directory = self._directory.joinpath(sub_dir)
        # directory = op.join(self._directory, sub_dir)
        try:
            directory.mkdir(parents=True, exist_ok=True)
            # os.makedirs(directory)
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise
        
        filename = op.join(sub_dir, name)
        full_path = self._directory.joinpath(filename)
        # full_path = op.join(self._directory, filename)
        return filename, full_path

    def remove(self, filename: str):
        """Remove a file given by `filename`.
        This method is cross-thread and cross-process safe. If an "error no
        entry" occurs, it is suppressed.
        :param str filename: relative path to file
        """
        # full_path = op.join(self._directory, filename)
        full_path = self._directory.joinpath(filename)
        try:
            # os.remove(full_path)
            full_path.unlink()

        except WindowsExceptionError:
            pass
        except OSError as error:
            if error.errno != errno.ENOENT:
                # ENOENT may occur if two caches attempt to delete the same
                # file at the same time.
                raise

    async def aremove(self, filename: str):
        """Remove a file given by `filename`.
        This method is cross-thread and cross-process safe. If an "error no
        entry" occurs, it is suppressed.
        :param str filename: relative path to file
        """
        # full_path = op.join(self._directory, filename)
        full_path = self._directory.joinpath(filename)
        try:
            # os.remove(full_path)
            await full_path.async_unlink()

        except WindowsExceptionError:
            pass
        except OSError as error:
            if error.errno != errno.ENOENT:
                # ENOENT may occur if two caches attempt to delete the same
                # file at the same time.
                raise




from __future__ import annotations

from io import IOBase
from lazyops.types import BaseModel, Field
from lazyops.utils import logger
from typing import Optional, Dict, Any, List, Union, Sequence, TYPE_CHECKING
from .types import SlackContext
from .configs import SlackSettings

if TYPE_CHECKING:
    from slack_sdk import WebClient
    from slack_sdk.web.async_client import AsyncWebClient

    from slack_sdk.models.attachments import Attachment
    from slack_sdk.models.blocks import Block


class SlackClient:

    verbose: Optional[bool] = False

    if TYPE_CHECKING:
        sapi: WebClient
        api: AsyncWebClient

    def __init__(
        self,
        token: Optional[str] = None,
        default_user: Optional[str] = None,
        default_channel: Optional[str] = None,
        username_mapping: Optional[Dict[str, Union[str, List[str]]]] = None,
        verbose: Optional[bool] = None,
        **kwargs
    ):
        """
        Slack Client
        """
        from lazyops.imports._slacksdk import resolve_slack_sdk
        resolve_slack_sdk(True)

        self.settings = SlackSettings()
        self.token = token or self.settings.bot_token
        self.disabled = self.token is None
        if self.token is None:
            logger.warning("No token provided. Disabling Slack client")
            return

        self._api: Optional['AsyncWebClient'] = None
        self._sapi: Optional['WebClient'] = None
        self._kwargs: Optional[Dict[str, Any]] = kwargs
        if verbose is not None: self.verbose = verbose
        self.logger = logger
        self.default_user = default_user or self.settings.default_user
        self.default_user_id: Optional[str] = None

        self.default_channel = default_channel or self.settings.default_channel
        self.default_channel_id: Optional[str] = None
        self.ctx: SlackContext = self.settings.load_context()
        if username_mapping: 
            self.ctx.username_mapping = username_mapping
            self.settings.save_context(self.ctx)

    @property
    def api(self) -> 'AsyncWebClient':
        """
        Returns the Async API
        """
        if self._api is None and not self.disabled:
            from slack_sdk.web.async_client import AsyncWebClient
            self._api = AsyncWebClient(token = self.token, **self._kwargs)
        return self._api
    
    @property
    def sapi(self) -> 'WebClient':
        """
        Returns the Sync API
        """
        if self._sapi is None and not self.disabled:
            from slack_sdk import WebClient
            self._sapi = WebClient(token = self.token, **self._kwargs)
        return self._sapi

    def get_users(self, **kwargs) -> List[Dict[str, Any]]:
        """
        Get users
        """
        resp = self.sapi.users_list(**kwargs)
        return resp["members"] if resp["ok"] else []

    async def aget_users(self, **kwargs) -> List[Dict[str, Any]]:
        """
        Get users
        """
        resp = await self.api.users_list(**kwargs)
        return resp["members"] if resp["ok"] else []

    def get_channels(self, types: Optional[List[str]] = None, include_users: Optional[bool] = True, **kwargs) -> List[Dict[str, Any]]:
        """
        Get channels
        """
        if not types:  types = ['public_channel','private_channel','im','mpim']
        if types and isinstance(types, list): types = ",".join(types)
        resp = self.sapi.conversations_list(types = types, **kwargs)
        channels =  resp["channels"] if resp["ok"] else []
        if include_users:
            resp = self.sapi.users_conversations(types = types, **kwargs)
            channels += resp["channels"] if resp["ok"] else []
        return channels
    
    async def aget_channels(self, types: Optional[List[str]] = None, include_users: Optional[bool] = True, **kwargs) -> List[Dict[str, Any]]:
        """
        Get channels
        """
        if not types:  types = ['public_channel','private_channel','im','mpim']
        if types and isinstance(types, list):
            types = ",".join(types)
        resp = await self.api.conversations_list(types = types, **kwargs)
        channels =  resp["channels"] if resp["ok"] else []
        if include_users:
            resp = await self.api.users_conversations(types = types, **kwargs)
            channels += resp["channels"] if resp["ok"] else []
        return channels

    @property
    def default_id(self) -> Optional[str]:
        """
        Default ID
        """
        if self.disabled: return
        return self.default_user_id or self.default_channel_id


    def configure_defaults(self):
        """
        Configure defaults
        """
        if self.default_user:
            self.default_user_id = self.ctx.lookup_id(self.default_user)
            if self.verbose: self.logger.info(f"Default user id: {self.default_user} -> {self.default_user_id}")
        if self.default_channel:
            self.default_channel_id = self.ctx.lookup_id(self.default_channel)
            if self.verbose: self.logger.info(f"Default channel id: {self.default_channel} -> {self.default_channel_id}")
        if self.default_id and self.verbose: self.logger.info(f"Default id: {self.default_id}")

    def init(self, overwrite: Optional[bool] = False):
        """
        Initialize
        """
        # Build Channels
        if self.disabled: return
        if not overwrite and self.ctx.initialized:
            self.configure_defaults()
            return

        users = self.get_users()
        for user in users:
            if aliases := self.ctx.username_mapping.get(user['name']):
                if isinstance(aliases, str): aliases = [aliases]
                for alias in aliases:
                    self.ctx.uids[user['id']] = alias
                    self.ctx.users[alias] =  user['id']
            else:
                self.ctx.uids[user['id']] = user['name']
            self.ctx.users[user['name']] =  user['id']


        channels = self.get_channels()
        for channel in channels:
            if not channel.get('name'): 
                self.ctx.channels[self.ctx.uids[channel['user']]] = channel['id']
                self.ctx.uids[channel['id']] = self.ctx.uids[channel['user']]
            else:
                self.ctx.channels[channel['name']] = channel['id']
                self.ctx.uids[channel['id']] = channel['name']
        
        self.ctx.initialized = True
        self.configure_defaults()
        self.settings.save_context(self.ctx)

    async def ainit(self, overwrite: Optional[bool] = False):
        """
        Initialize
        """
        # Build Channels
        if self.disabled: return
        if not overwrite and self.ctx.initialized:
            self.configure_defaults()
            return

        users = await self.aget_users()
        for user in users:
            if aliases := self.ctx.username_mapping.get(user['name']):
                if isinstance(aliases, str): aliases = [aliases]
                for alias in aliases:
                    self.ctx.uids[user['id']] = alias
                    self.ctx.users[alias] =  user['id']
            else:
                self.ctx.uids[user['id']] = user['name']
            self.ctx.users[user['name']] =  user['id']


        channels = await self.aget_channels()
        for channel in channels:
            if not channel.get('name'): 
                self.ctx.channels[self.ctx.uids[channel['user']]] = channel['id']
                self.ctx.uids[channel['id']] = self.ctx.uids[channel['user']]
            else:
                self.ctx.channels[channel['name']] = channel['id']
                self.ctx.uids[channel['id']] = channel['name']
        
        self.ctx.initialized = True
        self.configure_defaults()

    def lookup(self, user_or_channel: str) -> Optional[str]:
        """
        Return the ID of a user or channel
        """
        if self.disabled: return
        if user_or_channel in [
            self.default_id,
            self.default_channel_id,
        ]:
            return user_or_channel
        return self.ctx.lookup_id(user_or_channel) or user_or_channel


    def join_channel(self, channel_id: str, **kwargs):
        """
        Join a channel
        """
        if self.disabled: return
        if self.verbose: self.logger.info(f"Joining channel {channel_id}")
        self.sapi.conversations_join(channel = channel_id, **kwargs)

    async def ajoin_channel(self, channel_id: str, **kwargs):
        """
        Join a channel
        """
        if self.disabled: return
        if self.verbose: self.logger.info(f"Joining channel {channel_id}")
        await self.api.conversations_join(channel = channel_id, **kwargs)

    def message(
        self,
        message: Optional[str] = None,
        user_or_channel: Optional[str] = None,
        channel_id: Optional[str] = None,
        **kwargs
    ):
        """
        Send a message
        """
        if self.disabled: return
        if user_or_channel:
            channel_id = self.lookup(user_or_channel)
        elif channel_id is None:
            channel_id = self.default_id
        if not self.ctx.uids.get(channel_id): self.join_channel(channel_id)
        if channel_id is None:
            raise ValueError("No channel id provided")
        return self.sapi.chat_postMessage(channel = channel_id, text = message, **kwargs)
    

    async def amessage(
        self,
        message: str,
        user_or_channel: Optional[str] = None,
        channel_id: Optional[str] = None,
        **kwargs
    ):
        """
        Send a message
        """
        if self.disabled: return
        if user_or_channel:
            channel_id = self.lookup(user_or_channel)
        elif channel_id is None:
            channel_id = self.default_id
        if not self.ctx.uids.get(channel_id): await self.ajoin_channel(channel_id)
        if channel_id is None:
            raise ValueError("No channel id provided")
        return await self.api.chat_postMessage(channel = channel_id, text = message, **kwargs)
    

    def upload_file(
        self,
        file: Optional[Union[str, bytes, IOBase]] = None,
        content: Optional[str] = None,
        filename: Optional[str] = None,
        filetype: Optional[str] = None,
        initial_comment: Optional[str] = None,
        thread_ts: Optional[str] = None,
        title: Optional[str] = None,
        channels: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ):
        """
        Upload a file
        """
        if self.disabled: return
        channels = channels or [self.default_id]
        if not isinstance(channels, (list, tuple)): channels = [channels]
        if channels: channels = [self.lookup(channel) for channel in channels]
        return self.sapi.files_upload_v2(
            file = file, 
            content = content, 
            filename = filename, 
            filetype = filetype, 
            initial_comment = initial_comment, 
            thread_ts = thread_ts, 
            title = title, 
            channels = channels, 
            **kwargs
        )
    
    async def aupload_file(
        self,
        file: Optional[Union[str, bytes, IOBase]] = None,
        content: Optional[str] = None,
        filename: Optional[str] = None,
        filetype: Optional[str] = None,
        initial_comment: Optional[str] = None,
        thread_ts: Optional[str] = None,
        title: Optional[str] = None,
        channels: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ):
        """
        Upload a file
        """
        if self.disabled: return
        channels = channels or [self.default_id]
        if not isinstance(channels, (list, tuple)): channels = [channels]
        if channels: channels = [self.lookup(channel) for channel in channels]
        return await self.api.files_upload_v2(
            file = file, 
            content = content, 
            filename = filename, 
            filetype = filetype, 
            initial_comment = initial_comment, 
            thread_ts = thread_ts, 
            title = title, 
            channels = channels, 
            **kwargs
        )
    
    def temp_message(
        self,
        text: str,
        user: str,
        channel: Optional[str] = None,
        attachments: Optional[Sequence[Union[Dict, 'Attachment']]] = None,
        blocks: Optional[Sequence[Union[Dict, 'Block']]] = None,
        **kwargs,
    ):
        """
        Send a temporary message
        """
        if self.disabled: return
        user = self.lookup(user)
        channel = self.lookup(channel) if channel else self.default_id
        return self.sapi.chat_postEphemeral(
            text = text, 
            channel = channel, 
            user = user, 
            attachments = attachments, 
            blocks = blocks,
            **kwargs
        )
    
    async def atemp_message(
        self,
        text: str,
        user: str,
        channel: Optional[str] = None,
        attachments: Optional[Sequence[Union[Dict, 'Attachment']]] = None,
        blocks: Optional[Sequence[Union[Dict, 'Block']]] = None,
        **kwargs,
    ):
        """
        Send a temporary message
        """
        if self.disabled: return
        user = self.lookup(user)
        channel = self.lookup(channel) if channel else self.default_id
        return await self.api.chat_postEphemeral(
            text = text, 
            channel = channel, 
            user = user, 
            attachments = attachments, 
            blocks = blocks,
            **kwargs
        )


    def __getitem__(self, name_or_id: str) -> 'SlackProxy':
        """
        Return the ID of a user or channel
        """
        if self.disabled: return
        return SlackProxy(name = name_or_id, uid = self.lookup(name_or_id) or name_or_id, client = self)


class SlackProxy(BaseModel):

    name: str
    uid: str
    client: SlackClient


    def message(
        self,
        message: str,
        **kwargs
    ):
        """
        Send a message
        """
        return self.client.sapi.chat_postMessage(channel = self.uid, text = message, **kwargs)
    

    async def amessage(
        self,
        message: str,
        **kwargs
    ):
        """
        Send a message
        """
        return await self.client.api.chat_postMessage(channel = self.uid, text = message, **kwargs)
    
    def upload_file(
        self,
        file: Optional[Union[str, bytes, IOBase]] = None,
        content: Optional[str] = None,
        filename: Optional[str] = None,
        filetype: Optional[str] = None,
        initial_comment: Optional[str] = None,
        thread_ts: Optional[str] = None,
        title: Optional[str] = None,
        channels: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ):
        """
        Upload a file
        """
        channels = channels or [self.client.default_id]
        if not isinstance(channels, (list, tuple)): channels = [channels]
        if channels: 
            channels = [self.client.lookup(channel) for channel in channels]
        return self.client.sapi.files_upload_v2(
            file = file, 
            content = content, 
            filename = filename, 
            filetype = filetype, 
            initial_comment = initial_comment, 
            thread_ts = thread_ts, 
            title = title, 
            channels = channels, 
            **kwargs
        )
    

    async def aupload_file(
        self,
        file: Optional[Union[str, bytes, IOBase]] = None,
        content: Optional[str] = None,
        filename: Optional[str] = None,
        filetype: Optional[str] = None,
        initial_comment: Optional[str] = None,
        thread_ts: Optional[str] = None,
        title: Optional[str] = None,
        channels: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ):
        """
        Upload a file
        """
        channels = channels or [self.client.default_id]
        if not isinstance(channels, (list, tuple)): channels = [channels]
        if channels: 
            channels = [self.client.lookup(channel) for channel in channels]
        return await self.client.api.files_upload_v2(
            file = file, 
            content = content, 
            filename = filename, 
            filetype = filetype, 
            initial_comment = initial_comment, 
            thread_ts = thread_ts, 
            title = title, 
            channels = channels, 
            **kwargs
        )
    
    def temp_message(
        self,
        text: str,
        channel: str,
        attachments: Optional[Sequence[Union[Dict, 'Attachment']]] = None,
        blocks: Optional[Sequence[Union[Dict, 'Block']]] = None,
        **kwargs,
    ):
        """
        Send a temporary message
        """
        channel = self.client.lookup(channel)
        return self.client.sapi.chat_postEphemeral(
            text = text, 
            channel = channel, 
            user = self.uid, 
            attachments = attachments, 
            blocks = blocks,
            **kwargs
        )

    async def atemp_message(
        self,
        text: str,
        channel: str,
        attachments: Optional[Sequence[Union[Dict, 'Attachment']]] = None,
        blocks: Optional[Sequence[Union[Dict, 'Block']]] = None,
        **kwargs,
    ):
        """
        Send a temporary message
        """
        channel = self.client.lookup(channel)
        return await self.client.api.chat_postEphemeral(
            text = text, 
            channel = channel, 
            user = self.uid, 
            attachments = attachments, 
            blocks = blocks,
            **kwargs
        )




from .models import (
    LazyFile,
    LazyJson,
    LazyPickler,
    LazyIOBase,
    LazyIOText,
    LazyIOJson,
    LazyIOJsonLines,
    LazyIOPickle,
    LazyIOType,
    LazyHFModel,
    gfile,
    File,
    json,
    PathIO,
    PathIOLike,
)
from .async_ops import (
    async_open,
    aio_stat,
    aio_remove,
    aio_rename,
    aio_mkdir,
    stat_results
)

lazypickler = LazyPickler
lazypkler = LazyPickler
lazyjson = LazyJson

#lazyfile = LazyIOBase
lazyfile = LazyFile
lazyiojson = LazyIOJson
lazyjsonlines = LazyIOJsonLines
lazyjlines = LazyIOJsonLines
lazyjl = LazyIOJsonLines
lazypickle = LazyIOPickle
lazypkl = LazyIOPickle


__all__ = [
    'LazyFile',
    'LazyJson',
    'LazyPickler',
    'LazyIOBase',
    'LazyIOText',
    'LazyIOJson',
    'LazyIOJsonLines',
    'LazyIOPickle',
    'LazyIOType',
    'LazyHFModel',
    'lazypickler',
    'lazypkler',
    'lazyjson',
    'lazyfile',
    'lazyiojson',
    'lazyjsonlines',
    'lazyjlines',
    'lazyjl',
    'lazypickle',
    'lazypkl',
    'File',
    'gfile',
    'json',
    'PathIO',
    'PathIOLike',
    'async_open',
    'aio_stat',
    'aio_remove',
    'aio_rename',
    'aio_mkdir',
    'stat_results'
]




import io
import pickle
from typing import Union, Any
from pydantic import BaseModel
from pydantic.types import ByteSize

class Constant(tuple):
    """Pretty display of immutable constant."""

    def __new__(cls, name):
        return tuple.__new__(cls, (name,))

    def __repr__(self):
        return f'{self[0]}'


DBNAME = 'sqlcache.db'
ENOVAL = Constant('ENOVAL')
UNKNOWN = Constant('UNKNOWN')

MODE_NONE = 0
MODE_RAW = 1
MODE_BINARY = 2
MODE_TEXT = 3
MODE_PICKLE = 4

KeyT = Union[str, int, float, bytes, None]
ValueT = Union[str, int, float, bytes, None, io.BytesIO, Any]

class SizeModel(BaseModel):
    size: ByteSize



DEFAULT_SETTINGS = {
    'statistics': 0,  # False
    'tag_index': 0,  # False
    'eviction_policy': 'least-recently-stored',
    'size_limit': 2**30,  # 1gb
    'cull_limit': 10,
    'sqlite_auto_vacuum': 1,  # FULL
    'sqlite_cache_size': 2**13,  # 8,192 pages
    'sqlite_journal_mode': 'wal',
    'sqlite_mmap_size': 2**26,  # 64mb
    'sqlite_synchronous': 1,  # NORMAL
    'disk_min_file_size': 2**15,  # 32kb
    'disk_pickle_protocol': pickle.HIGHEST_PROTOCOL,
}

OPTIMIZED_SETTINGS = {
    'fast': {
        'min_file_size': SizeModel(size = '32KiB').size,
        'size_limit': SizeModel(size = '4GiB').size,
        'mmap_size': SizeModel(size = '512MiB').size,
        'eviction_policy': 'least-recently-stored',
        'cache_size': 8192,
        'compression_level': 0,
    },
    'standard': {
        'min_file_size': SizeModel(size = '64KiB').size,
        'size_limit': SizeModel(size = '8GiB').size,
        'mmap_size': SizeModel(size = '1024MiB').size,
        'eviction_policy': 'least-recently-stored',
        'cache_size': 8192,
        'compression_level': 5,
    },
    'optimized': {
        'min_file_size': SizeModel(size = '128KiB').size,
        'size_limit': SizeModel(size = '16GiB').size,
        'mmap_size': SizeModel(size = '2048MiB').size,
        'eviction_policy': 'least-recently-stored',
        'sqlite_cache_size': 8192,
        'compression_level': 2,
    },
    'cache': {
        'min_file_size': SizeModel(size = '256KiB').size,
        'size_limit': SizeModel(size = '32GiB').size,
        'mmap_size': SizeModel(size = '4GiB').size,
        'eviction_policy': 'least-recently-used',
        'cache_size': 8192 * 2,
        'compression_level': 8,
    },
    'cache_large': {
        'min_file_size': SizeModel(size = '512KiB').size,
        'size_limit': SizeModel(size = '64GiB').size,
        'mmap_size': SizeModel(size = '8GiB').size,
        'eviction_policy': 'least-recently-used',
        'cache_size': 8192 * 4,
        'compression_level': 10,
    },
}

METADATA = {
    'count': 0,
    'size': 0,
    'hits': 0,
    'misses': 0,
}

EVICTION_POLICY = {
    'none': {
        'init': None,
        'get': None,
        'cull': None,
    },
    'least-recently-stored': {
        'init': (
            'CREATE INDEX IF NOT EXISTS Cache_store_time ON'
            ' Cache (store_time)'
        ),
        'get': None,
        'cull': 'SELECT {fields} FROM Cache ORDER BY store_time LIMIT ?',
    },
    'least-recently-used': {
        'init': (
            'CREATE INDEX IF NOT EXISTS Cache_access_time ON'
            ' Cache (access_time)'
        ),
        'get': 'access_time = {now}',
        'cull': 'SELECT {fields} FROM Cache ORDER BY access_time LIMIT ?',
    },
    'least-frequently-used': {
        'init': (
            'CREATE INDEX IF NOT EXISTS Cache_access_count ON'
            ' Cache (access_count)'
        ),
        'get': 'access_count = access_count + 1',
        'cull': 'SELECT {fields} FROM Cache ORDER BY access_count LIMIT ?',
    },
}

_EVICTION_POLICY = {
    'none': {
        'init': None,
        'get': None,
        'cull': None,
    },
    'least-recently-stored': {
        'init': (
            'CREATE INDEX IF NOT EXISTS Cache_store_time ON'
            ' Cache (store_time)'
        ),
        'get': None,
        'cull': 'SELECT {fields} FROM Cache ORDER BY store_time LIMIT ?',
    },
    'least-recently-used': {
        'init': (
            'CREATE INDEX IF NOT EXISTS Cache_access_time ON'
            ' Cache (access_time)'
        ),
        'get': 'access_time = {now}',
        'cull': 'SELECT {fields} FROM Cache ORDER BY access_time LIMIT ?',
    },
    'least-frequently-used': {
        'init': (
            'CREATE INDEX IF NOT EXISTS Cache_access_count ON'
            ' Cache (access_count)'
        ),
        'get': 'access_count = access_count + 1',
        'cull': 'SELECT {fields} FROM Cache ORDER BY access_count LIMIT ?',
    },
}



try: 
    class WindowsExceptionError(WindowsError):
        pass
except:
    class WindowsExceptionError(Exception):
        "Windows error place-holder on platforms without support."



from __future__ import annotations

"""
Rework v3 of the logger

- Now imported from `lazyops.libs.logging` 
"""

from lazyops.libs.logging import (
    DEFAULT_STATUS_COLORS,
    QUEUE_STATUS_COLORS,
    STATUS_COLOR,
    FALLBACK_STATUS_COLOR,
    DEFAULT_FUNCTION_COLOR,
    DEFAULT_CLASS_COLOR,
    RESET_COLOR,
    LOGLEVEL_MAPPING,
    REVERSE_LOGLEVEL_MAPPING,
    COLORED_MESSAGE_MAP,

    NullLogger,
    Logger,

    InterceptHandler,
    LoggerFormatter,

    register_logger_module,
    is_registered_logger_module,
    create_default_logger,

    get_logger,
    default_logger,
    null_logger,
    logger,

    change_logger_level,
    add_api_log_filters,
)


from __future__ import annotations

import os
import signal
import asyncio
import contextlib

from typing import Optional, List, TypeVar, Callable, Dict, Any, TYPE_CHECKING

if TYPE_CHECKING:
    with contextlib.suppress(ImportError):
        from aiokeydb.types.worker import Worker


_WorkerCtx: Dict[str, Dict[str, List['Worker']]] = {}


async def stop_worker(
    loop: asyncio.BaseEventLoop,
    worker: 'Worker',
    signame: str,
    verbose: Optional[bool] = True,
):
    """
    Stops the worker
    """
    from lazyops.utils import logger
    await worker.stop()
    if verbose: logger.warning(f'Worker {worker.name}: Received exit signal {signame} for ...')
    tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
    [task.cancel() for task in tasks]
    if verbose: logger.warning(f'Worker {worker.name}: Cancelling {len(tasks)} outstanding tasks')
    loop.stop()
    if verbose: logger.warning(f'Worker {worker.name}: Loop stopped')


def spawn_new_worker(
    name: str, 
    index: Optional[int] = None, 
    kind: Optional[str] = None,
    config: Optional[Dict[str, Any]] = None, 
    verbose: Optional[bool] = True, 
    is_primary_worker: Optional[bool] = True, 
    settings_config_func: Optional[Callable] = None, # should configure the settings
    get_worker_func: Optional[Callable] = None,
    **kwargs
):
    """
    Spawns a new worker
    """
    global _WorkerCtx
    assert get_worker_func is not None, "get_worker_func must be defined"
    os.environ['IS_WORKER_PROCESS'] = 'True'
    
    if index is None: index = 0
    if kind is None: kind = 'default'
    if settings_config_func is not None:
        settings_config_func(
            name = name,
            index = index,
            kind = kind,
            config = config,
            verbose = verbose,
            is_primary_worker = is_primary_worker,
            **kwargs
        )
    
    worker: 'Worker' = get_worker_func(
        name = name,
        index = index,
        kind = kind,
        config = config,
        verbose = verbose,
        is_primary_worker = is_primary_worker,
        **kwargs
    )
    worker.name = f"{worker.name}-{index}"
    if worker.is_leader_process is None:
        worker.is_leader_process = is_primary_worker
    loop = asyncio.new_event_loop()
    for signame in ('SIGINT', 'SIGTERM'):
        loop.add_signal_handler(getattr(signal, signame), lambda signame=signame: asyncio.create_task(stop_worker(loop, worker, signame, verbose = verbose)))
        
    if kind not in _WorkerCtx: _WorkerCtx[kind] = {}
    if name not in _WorkerCtx[kind]: _WorkerCtx[kind][name] = []
    _WorkerCtx[kind][name].append(worker)
    loop.run_until_complete(worker.start())

    

async def terminate_worker(name: str, kind: Optional[str] = None):
    """
    Terminates a worker
    """
    global _WorkerCtx
    if kind is None: kind = 'default'
    if kind not in _WorkerCtx: return
    if name not in _WorkerCtx[kind]: return
    for worker in _WorkerCtx[kind][name]:
        await worker.stop()
        _WorkerCtx[kind][name].remove(worker)
    





from __future__ import annotations

import zlib
from .base import BaseCompression, logger
from typing import Optional

class ZlibCompression(BaseCompression):
    name: str = "zlib"
    compression_level: Optional[int] = -1

    def validate_compression_level(self):
        """
        Validates the compression level
        """
        assert self.compression_level == -1 or self.compression_level in range(10), "Compression level must be between 0 and 9 or -1"


    def compress(self, data: bytes, level: Optional[int] = None, **kwargs) -> bytes:
        """
        Compresses the data
        """
        if level is None: level = self.compression_level
        return zlib.compress(data, level = level)

    def decompress(self, data: bytes, **kwargs) -> bytes:
        """
        Decompresses the data
        """
        _kwargs = self._decompression_kwargs.copy()
        if kwargs: _kwargs.update(kwargs)
        return zlib.decompress(data, **_kwargs)
    






from lazyops.imports._aiohttpx import aiohttpx
from lazyops.types.classprops import lazyproperty

class ClientError(Exception):
    client_name: str = 'APIClient'

    def __init__(
        self, 
        response: 'aiohttpx.Response'
    ):
        self.response = response

    @lazyproperty
    def url(self):
        return self.response.url
    
    @lazyproperty
    def status_code(self):
        return self.response.status_code
    
    @lazyproperty
    def headers(self):
        return self.response.headers
    
    @lazyproperty
    def payload(self):
        try:
            return self.response.json()
        except Exception:
            return self.response.text


    def __str__(self):
        return f"[{self.client_name}] url: {self.url}, status_code: {self.status_code}, payload: {self.payload}"


def fatal_exception(exc):
    if isinstance(exc, ClientError):
        # retry on no sessions available.
        # retry on overloaded
        # if exc.status_code in [500, 503]:
        #     return False
        # return exc.status_code >= 400
        return exc.status_code == 503 or exc.status_code >= 400
    else:
        # retry on all other errors (eg. network)
        return False





from lazyops.types.common import StrEnum, UpperStrEnum

class New(StrEnum):
    """
    The job status
    """
    NEW = 'new'
    START = 'start'
    STAGING = 'staging'
    FINAL = 'final'

    def __contains__(self, other) -> bool:
        __key = other.lower() if isinstance(other, str) else other
        print(__key)
        return super().__contains__(__key)

# this returns false
x = (
    'NEW' in [
        New.NEW,
    ]
)
print(x)

# This returns True
print('new' in New.NEW)



__all__ = ["Database", "DatabaseContent", "Role", "Schema"]

from lazyops.libs.dbinit.entities.database import Database
from lazyops.libs.dbinit.entities.database_content import DatabaseContent
from lazyops.libs.dbinit.entities.role import Role
from lazyops.libs.dbinit.entities.schema import Schema



from __future__ import absolute_import

from lazyops.types.common import *

from lazyops.types.models import BaseModel, BaseSettings, Schema, Field, validator, root_validator, pre_root_validator, PYD_VERSION, schema_extra
from lazyops.types.models import UrlModel, BaseDBUrl, PostgresDB, RedisDB, NatsDB
from lazyops.types.classprops import lazyproperty, classproperty, cached_classproperty

try:
    from pydantic.dataclasses import dataclass
except ImportError:
    from dataclasses import dataclass

# from lazyops.types.resources import BaseResource, ResourceType, ResponseResource, ResponseResourceType


import uuid
import json
import typing
import codecs
import hashlib
import datetime
import contextlib
import dataclasses
from enum import Enum

try:
    import numpy as np
    
except ImportError:
    np = None



# try:
#     import ruamel.yaml.scalarstring as yaml
# except ImportError:
#     yaml = None

# Borrowed from httpx
# Null bytes; no need to recreate these on each call to guess_json_utf
_null = b"\x00"
_null2 = _null * 2
_null3 = _null * 3

def guess_json_utf(data: bytes) -> typing.Optional[str]:
    # JSON always starts with two ASCII characters, so detection is as
    # easy as counting the nulls and from their location and count
    # determine the encoding. Also detect a BOM, if present.
    sample = data[:4]
    if sample in (codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE):
        return "utf-32"  # BOM included
    if sample[:3] == codecs.BOM_UTF8:
        return "utf-8-sig"  # BOM included, MS style (discouraged)
    if sample[:2] in (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE):
        return "utf-16"  # BOM included
    nullcount = sample.count(_null)
    if nullcount == 0:
        return "utf-8"
    if nullcount == 2:
        if sample[::2] == _null2:  # 1st and 3rd are null
            return "utf-16-be"
        if sample[1::2] == _null2:  # 2nd and 4th are null
            return "utf-16-le"
    elif nullcount == 3:
        if sample[:3] == _null3:
            return "utf-32-be"
        if sample[1:] == _null3:
            return "utf-32-le"
    return None


def parse_list_str(
    line: typing.Optional[typing.Union[typing.List[str], str]],
    default: typing.Optional[typing.List[str]] = None,
    seperators: typing.Optional[typing.List[str]] = None,
) -> typing.Optional[typing.List[str]]:
    """
    Try to parse a string as a list of strings

    Args:
        line (typing.Optional[typing.Union[typing.List[str], str]]): [description]
        default (typing.Optional[typing.List[str]], optional): [description]. Defaults to None.
        seperators (typing.Optional[typing.List[str]], optional): [description]. Defaults to None.
    
    """
    if line is None: return default
    if seperators is None: seperators = [',', '|', ';']
    if isinstance(line, list): return line
    if isinstance(line, str):
        if '[' in line and ']' in line:
            if '"' in line or "'" in line:
                try:
                    line = json.loads(line)
                    return line
                except Exception: line = line.replace("'", '').replace('"', '')
            line = line.replace('[', '').replace(']', '')
        for seperator in seperators:
            if seperator in line:
                return line.split(seperator)
        line = [line]
    return line
    


def object_serializer(obj: typing.Any) -> typing.Any:
    if isinstance(obj, dict):
        return {k: object_serializer(v) for k, v in obj.items()}

    if isinstance(obj, list):
        return [object_serializer(v) for v in obj]    

    if isinstance(obj, bytes):
        return obj.decode('utf-8')
    
    if isinstance(obj, (str, int, float, bool, type(None))):
        return obj

    if hasattr(obj, 'dict'): # test for pydantic models
        return obj.dict()
    
    if dataclasses.is_dataclass(obj):
        return dataclasses.asdict(obj)

    # Custom dict methods
    if hasattr(obj, 'to_dict'):
        return obj.to_dict()
    
    if hasattr(obj, 'todict'):
        return obj.todict()
    
    if hasattr(obj, 'to_json'):
        return obj.to_json()
    
    if hasattr(obj, 'tojson'):
        return obj.tojson()
    
    if hasattr(obj, 'toJson'):
        return obj.toJson()

    if hasattr(obj, 'json'):
        return obj.json()
    
    if hasattr(obj, 'encode'):
        return obj.encode()

    if hasattr(obj, 'get_secret_value'):
        return obj.get_secret_value()
    
    if hasattr(obj, 'as_posix'):
        return obj.as_posix()
    
    if hasattr(obj, "numpy"):  # Checks for TF tensors without needing the import
        return obj.numpy().tolist()
    
    if hasattr(obj, 'tolist'): # Checks for torch tensors without importing
        return obj.tolist()
    
    # Convert pd datetimes to isoformat
    if isinstance(obj, (datetime.date, datetime.datetime)) or hasattr(obj, 'isoformat'):
        return obj.isoformat()
    
    # Convert UUIDs
    if isinstance(obj, uuid.UUID):
        return str(obj)
    
    if isinstance(obj, Enum): #  hasattr(obj, 'value'):
        return obj.value

    if np is not None:
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64)):
            return int(obj)
        
        if isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
            return float(obj)

    if isinstance(obj, object):
        with contextlib.suppress(Exception):
            return {k: object_serializer(v) for k, v in obj.__dict__.items()}
    
    
    # Try to convert to a primitive type
    with contextlib.suppress(Exception):
        return int(obj)
    
    with contextlib.suppress(Exception):
        return float(obj)
    
    with contextlib.suppress(Exception):
        return str(obj)


    raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


def object_deserializer(obj: typing.Dict) -> typing.Dict:
    results = {}

    for key, value in obj.items():
        
        if isinstance(value, bytes):
            encoding = guess_json_utf(value)
            results[key] = value.decode(encoding) if encoding is not None else value
            continue

        if isinstance(value, dict):
            results[key] = object_deserializer(value)
            continue

        for dt_key in {
            'created',
            'updated',
            'modified',
            'timestamp',
            'date',
        }:
            if dt_key in key:
                if isinstance(value, str):
                    with contextlib.suppress(Exception):
                        results[key] = datetime.datetime.strptime(value, '%a, %d %b %Y %H:%M:%S GMT')
                        continue
                    with contextlib.suppress(Exception):
                        results[key] = datetime.datetime.fromisoformat(value)
                        continue
                
                with contextlib.suppress(Exception):
                    results[key] = datetime.datetime.fromtimestamp(value, tz = datetime.timezone.utc)
                    continue


        results[key] = value
    return results


if typing.TYPE_CHECKING:
    from lazyops.types import BaseModel

from .lazy import lazy_import, get_obj_class_name

def object_model_serializer(obj: typing.Union['BaseModel', typing.Any]) -> typing.Any:
    """
    Hooks for the object serializer for BaseModels
    """
    if not hasattr(obj, "Config") and not hasattr(obj, "dict"):
        return object_serializer(obj)
    return {
        "__jsontype__": "model",
        "__model__": get_obj_class_name(obj),
        "__data__": obj.dict(),
    }


def object_model_deserializer(obj: typing.Any) -> typing.Union['BaseModel', typing.Any]:
    """
    Hooks for the object deserializer for BaseModels
    """
    if not isinstance(obj, dict): return obj
    if any(key not in obj for key in ["__jsontype__", "__model__", "__data__"]):
        return object_deserializer(obj)

    
    model = lazy_import(obj["__model__"])
    return model(**obj["__data__"])


class ObjectEncoder(json.JSONEncoder):
    
    def default(self, obj: typing.Any):   # pylint: disable=arguments-differ,method-hidden
        with contextlib.suppress(Exception):
           return object_serializer(obj)
        # try:
        #     return object_serializer(obj)
        # except TypeError:
        #     print(f"Object of type {type(obj)} is not JSON serializable: {obj}")
        return json.JSONEncoder.default(self, obj)

class ObjectDecoder(json.JSONDecoder):
    
    def __init__(self, *args, object_hook: typing.Optional[typing.Callable] = None, **kwargs):
        object_hook = object_hook or object_deserializer
        super().__init__(*args, object_hook = object_hook, **kwargs)

class ObjectModelEncoder(json.JSONEncoder):
    """
    Object Model Encoder
    """
    def default(self, obj: typing.Any):   # pylint: disable=arguments-differ,method-hidden
        with contextlib.suppress(Exception):
            return object_model_serializer(obj)
        return json.JSONEncoder.default(self, obj)
        
class ObjectModelDecoder(json.JSONDecoder):
    """
    Object Model Decoder
    """
    def __init__(self, *args, object_hook: typing.Optional[typing.Callable] = None, **kwargs):
        object_hook = object_hook or object_model_deserializer
        super().__init__(*args, object_hook = object_hook, **kwargs)


class Json:

    @staticmethod
    def dumps(
        obj: typing.Dict[typing.Any, typing.Any], 
        *args, 
        default: typing.Dict[typing.Any, typing.Any] = None, 
        cls: typing.Type[json.JSONEncoder] = ObjectEncoder,
        _fallback_method: typing.Optional[typing.Callable] = None,
        **kwargs
    ) -> str:
        try:
            return json.dumps(obj, *args, default = default, cls = cls, **kwargs)
        except Exception as e:
            if _fallback_method is not None:
                return _fallback_method(obj, *args, default = default, **kwargs)
            raise e

    @staticmethod
    def loads(
        data: typing.Union[str, bytes], 
        *args, 
        _fallback_method: typing.Optional[typing.Callable] = None,
        **kwargs
    ) -> typing.Union[typing.Dict[typing.Any, typing.Any], typing.List[str]]:
        try:
            return json.loads(data, *args, **kwargs)
        except json.JSONDecodeError as e:
            if _fallback_method is not None:
                return _fallback_method(data, *args, **kwargs)
            raise e

class JsonModelSerializer:

    """
    Encoder and Decoder for Pydantic Models
    for optimal performance in deep serialization
    """

    @staticmethod
    def dumps(
        obj: typing.Dict[typing.Any, typing.Any], 
        *args, 
        default: typing.Dict[typing.Any, typing.Any] = None, 
        cls: typing.Type[json.JSONEncoder] = ObjectModelEncoder,
        _fallback_method: typing.Optional[typing.Callable] = None,
        **kwargs
    ) -> str:
        """
        Serializes a dict into a JSON string using the ObjectModelEncoder
        """
        try:
            return json.dumps(obj, *args, default = default, cls = cls, **kwargs)
        except Exception as e:
            if _fallback_method is not None:
                return _fallback_method(obj, *args, default = default, **kwargs)
            raise e

    @staticmethod
    def loads(
        data: typing.Union[str, bytes], 
        *args, 
        cls: typing.Type[json.JSONDecoder] = ObjectModelDecoder,
        _fallback_method: typing.Optional[typing.Callable] = None,
        **kwargs
    ) -> typing.Union[typing.Dict[typing.Any, typing.Any], typing.List[str]]:
        """
        Loads a JSON string into a dict using the ObjectModelDecoder
        """
        try:
            return json.loads(data, *args, cls = cls, **kwargs)
        except json.JSONDecodeError as e:
            if _fallback_method is not None:
                return _fallback_method(data, *args, **kwargs)
            raise e

# Add simdjson support if available
try:
    import simdjson

    _parser = simdjson.Parser()

    class SimdJson:

        """
        JSON Encoder and Decoder using simdjson
        """
        parser = _parser

        @staticmethod
        def dumps(
            obj: typing.Dict[typing.Any, typing.Any], 
            *args, 
            default: typing.Dict[typing.Any, typing.Any] = None, 
            cls: typing.Type[json.JSONEncoder] = None,
            _fallback_method: typing.Optional[typing.Callable] = None,
            **kwargs
        ) -> str:
            """
            Serializes a dict into a JSON string
            """
            try:
                return simdjson.dumps(obj, *args, default = default, cls = cls, **kwargs)
            except Exception as e:
                if _fallback_method is not None:
                    return _fallback_method(obj, *args, default = default, **kwargs)
                raise e

        @staticmethod
        def loads(
            data: typing.Union[str, bytes], 
            *args, 
            object_hook: typing.Optional[typing.Callable] = None,
            recursive: typing.Optional[bool] = True,
            _raw: typing.Optional[bool] = False,
            _fallback_method: typing.Optional[typing.Callable] = None,
            **kwargs
        ) -> typing.Union[typing.Dict[typing.Any, typing.Any], typing.List[str], simdjson.Object, simdjson.Array]:
            """
            Loads a JSON string into a dict using the ObjectModelDecoder
            """
            try:
                value = _parser.parse(data, recursive)
                return value if _raw or not object_hook else object_hook(value)
            
            except Exception as e:
                if _fallback_method is not None:
                    return _fallback_method(data, *args, **kwargs)
                raise e


    class SimdJsonModelSerializer:

        """
        JSON Encoder and Decoder using simdjson
        """

        parser = _parser

        @staticmethod
        def dumps(
            obj: typing.Dict[typing.Any, typing.Any], 
            *args, 
            default: typing.Dict[typing.Any, typing.Any] = None, 
            cls: typing.Type[json.JSONEncoder] = ObjectModelEncoder,
            _fallback_method: typing.Optional[typing.Callable] = None,
            **kwargs
        ) -> str:
            """
            Serializes a dict into a JSON string using the ObjectModelEncoder
            """
            try:
                return simdjson.dumps(obj, *args, default = default, cls = cls, **kwargs)
            except Exception as e:
                if _fallback_method is not None:
                    return _fallback_method(obj, *args, default = default, **kwargs)
                raise e

        @staticmethod
        def loads(
            data: typing.Union[str, bytes], 
            *args, 
            object_hook: typing.Optional[typing.Callable] = object_model_deserializer,
            recursive: typing.Optional[bool] = True,
            _raw: typing.Optional[bool] = False,
            _fallback_method: typing.Optional[typing.Callable] = None,
            **kwargs
        ) -> typing.Union[typing.Dict[typing.Any, typing.Any], typing.List[str], 'BaseModel', simdjson.Object, simdjson.Array]:
            """
            Loads a JSON string into a dict using the ObjectModelDecoder
            """
            try:
                value = _parser.parse(data, recursive)
                return value if _raw or not object_hook else object_hook(value)
            
            except Exception as e:
                if _fallback_method is not None:
                    return _fallback_method(data, *args, **kwargs)
                raise e



except ImportError:
    _parser = None

    SimdJson = Json
    SimdJsonModelSerializer = JsonModelSerializer



"""
Hashing Functions
"""


def create_hash_key(
    args: typing.Optional[tuple] = None, 
    kwargs: typing.Optional[dict] = None, 
    typed: typing.Optional[bool] = False,
    key_base: typing.Optional[tuple] = None,
    exclude: typing.Optional[typing.List[str]] = None,
    hashfunc: typing.Optional[str] = 'md5',
    separator: typing.Optional[str] = ':',
):
    """
    Create hash key out of function arguments.
    :param tuple base: base of key
    :param tuple args: function arguments
    :param dict kwargs: function keyword arguments
    :param bool typed: include types in cache key
    :return: cache key tuple
    """

    hash_key = key_base or ()
    if args: hash_key += args
    if kwargs:
        if exclude: kwargs = {k: v for k, v in kwargs.items() if k not in exclude}
        sorted_items = sorted(kwargs.items())
        for item in sorted_items:
            hash_key += item

    if typed:
        hash_key += tuple(type(arg) for arg in args)
        if kwargs: hash_key += tuple(type(value) for _, value in sorted_items)

    cache_key = f'{separator}'.join(str(k) for k in hash_key)
    func = getattr(hashlib, hashfunc)
    return func(cache_key.encode()).hexdigest()




"""
Base Types from Typing
"""

__all__ = (
    'PathLike',
    'TYPE_CHECKING',
    'List', 'Dict', 'AnyStr', 'Any', 'Set',
    'Optional', 'Union', 'Tuple', 'Mapping', 'Sequence', 'TypeVar', 'Type',
    'Callable', 'Coroutine', 'Generator', 'AsyncGenerator', 'IO', 'Iterable', 'Iterator', 'AsyncIterator',
    'cast', 'overload',
    'Final', 'Literal',
    'Data', 'AnyMany', 'TextMany', 'TextList',
    'DictList', 'DictMany', 'DictAny', 'DictAny',
    'aobject',
    'StrEnum',
    'UpperStrEnum',
)

import sys

from os import PathLike
from typing import TYPE_CHECKING
from typing import List, Dict, AnyStr, Any, Set
from typing import Optional, Union, Tuple, Mapping, Sequence, TypeVar, Type
from typing import Callable, Coroutine, Generator, AsyncGenerator, IO, Iterable, Iterator, AsyncIterator
from typing import cast, overload


if sys.version_info >= (3, 8):
    from typing import Final, Literal
else:
    from typing_extensions import Final, Literal

Data = TypeVar('Data', str, List[str], Dict[str, Union[str, List[str]]])
AnyMany = TypeVar('AnyMany', Any, List[Any])

TextMany = TypeVar('TextMany', str, List[str])
TextList = List[str]

DictList = List[Dict[str, Any]]
DictMany = TypeVar('DictMany', Dict[str, Any], List[Dict[str, Any]])
DictAny = Dict[str, Any]
DictText = Dict[str, str]
NoneType = type(None)



class aobject(object):
    """Inheriting this class allows you to define an async __init__.

    So you can create objects by doing something like `await MyClass(params)`
    """
    async def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        await instance.__init__(*args, **kwargs)
        return instance

    async def __init__(self, *args, **kwargs):
        pass

# borrowed from 
# https://github.com/pydantic/pydantic/issues/1439
def construct_union(modules: List):  
    # modules is a list of class constructors and has to be accessible in the current global scope
    eval_string = ', '.join([f'{i.__module__}.{i.__name__}' for i in modules])
    return eval(f'Union[{eval_string}]')

# def construct_union(modules: List):
#     return Union[tuple(modules)]

from enum import Enum, EnumMeta
from functools import singledispatchmethod

class StrEnumMeta(EnumMeta):

    @singledispatchmethod
    def __getitem__(self, key):
        return super().__getitem__(key)

    @__getitem__.register
    def _(self, index: int):
        return list(self)[index]

class StrEnum(str, Enum, metaclass=StrEnumMeta):
    """
    StrEnum is a string enum that allows for case-insensitive comparisons
    """

    def __eq__(self, other: Any) -> bool:
        return self.value.lower() == other.lower() if \
            isinstance(other, str) else \
                super().__eq__(other)
    
    def __ne__(self, other: Any) -> bool:
        return self.value.lower() != other.lower() if \
            isinstance(other, str) else \
                super().__ne__(other)

    def __str__(self) -> str:
        return str.__str__(self)
    
    def __hash__(self) -> int:
        return id(self)


class UpperStrEnum(StrEnum):
    """
    UpperStrEnum is a string enum that allows for case-insensitive comparisons
    """

    def __eq__(self, other: Any) -> bool:
        return self.value.upper() == other.upper() if \
            isinstance(other, str) else \
                super().__eq__(other)
    
    def __ne__(self, other: Any) -> bool:
        return self.value.upper() != other.upper() if \
            isinstance(other, str) else \
                super().__ne__(other)
    

    def __str__(self) -> str:
        return str.__str__(self)
    
    def __hash__(self) -> int:
        return id(self)



if TYPE_CHECKING:
    ListStr = List[str]

else:

    class ListStr(list):
        """
        Returns List[str] by splitting on delimiter ','
        """

        @classmethod
        def __get_validators__(cls):
            yield cls.validate

        @classmethod
        def validate(cls, value: Optional[Any]) -> Optional[List[str]]:
            """
            Validates the value and returns a List[str]
            """
            if v is None: return None
            from lazyops.utils.serialization import parse_list_str
            return parse_list_str(value)



        



import traceback

from fastapi.exceptions import HTTPException
from lazyops.utils.logs import logger

from typing import Optional


class ORMException(HTTPException):
    """
    Base ORM Exception
    """
    
    base: Optional[str] = None
    concat_detail: Optional[bool] = True
    verbose: Optional[int] = None
    default_status_code: Optional[int] = 500
    log_error: Optional[bool] = True # Always log the error message

    def __init__(self, detail: str = None, error: Optional[Exception] = None, status_code: Optional[int] = None, **kwargs):
        """
        Constructor
        """
        message = ""
        if not detail or self.concat_detail:
            message += f"{self.base}: " or ''
        if detail: message += detail
        if error and self.verbose is not None:
            if self.verbose >= 1:
                message += f"\nError: {error}"
            elif self.verbose <= 5:
                message += f"\nError: {error}\nTraceback: {traceback.format_exc()}"
        status_code = status_code or self.default_status_code
        super().__init__(status_code=status_code, detail=message, **kwargs)
        if self.log_error:
            self.display(message, status_code)
    
    def display(self, message: str, status_code: int):
        """
        Displays the error
        """
        logger.error(f'[{self.__class__.__name__} - {status_code}]: {message}')


class NoItemFoundException(ORMException):
    """
    No Item Found Exception
    """
    base = "No Item Found"
    default_status_code = 404
    
class MissingItemsException(ORMException):
    """
    Missing Items Exception
    """
    base = "Not all Items were found for"
    default_status_code = 404
    

class NoResultsFoundException(ORMException):
    """
    No Results Found Exception
    """
    base = "No Results Found"
    default_status_code = 404

class InvalidModelException(ORMException):
    """
    Invalid Model Exception
    """
    base = "Invalid Model"
    default_status_code = 400

class InvalidOperationException(ORMException):
    """
    Invalid Operation Exception
    """
    base = "Invalid Operation"
    default_status_code = 400
    
class InvalidParameterException(ORMException):
    """
    Invalid Parameter Exception
    """
    base = "Invalid Parameter"
    default_status_code = 400




"""
Import Handler for `aiofiles`
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import aiofiles
    _aiofiles_available = True
except ImportError:
    aiofiles = object
    _aiofiles_available = False

def resolve_aiofiles(
    required: bool = False,
):
    """
    Ensures that `aiofiles` is available
    """
    global aiofiles, _aiofiles_available
    if not _aiofiles_available:
        resolve_missing('aiofiles', required = required)
        import aiofiles
        _aiofiles_available = True
        globals()['aiofiles'] = aiofiles


def require_aiofiles(
    required: bool = False,
):
    """
    Wrapper for `resolve_aiofiles` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_aiofiles, 
            func = func, 
            required = required
        )
    return decorator



"""
Import Handler for aiokeydb
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

_min_version = 'v0.1.0'

try:
    import aiokeydb
    _aiokeydb_available = True
except ImportError:
    aiokeydb = object
    _aiokeydb_available = False

def resolve_aiokeydb(
    required: bool = False,
    version: str = _min_version,
):
    """
    Ensures that `aiokeydb` is available
    """
    global aiokeydb, _aiokeydb_available
    if not _aiokeydb_available:
        resolve_missing(f'aiokeydb=={version}', required = required)
        import aiokeydb
        _aiokeydb_available = True
        globals()['aiokeydb'] = aiokeydb


def require_aiokeydb(
    required: bool = False,
    version: str = _min_version,
):
    """
    Wrapper for `resolve_aiokeydb` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_aiokeydb, 
            func = func, 
            required = required,
            version = version,
        )
    return decorator


from __future__ import annotations

"""
Base Serializers
"""

import abc
import zlib
import hashlib
from lazyops.types import BaseModel
from lazyops.utils.logs import logger
# from lazyops.utils.pooler import ThreadPoolV2 as ThreadPooler
from lazyops.utils.pooler import ThreadPooler
from typing import Any, Optional, Union, Dict, TypeVar, TYPE_CHECKING

try:
    import xxhash
    _xxhash_available = True
except ImportError:
    _xxhash_available = False

if TYPE_CHECKING:
    from ..compression import CompressionT


SchemaType = TypeVar("SchemaType", bound=BaseModel)
ObjectValue = Union[SchemaType, Dict, Any]

def create_obj_hash(obj: ObjectValue) -> str:
    """
    Creates a hash for the object
    """
    if isinstance(obj, BaseModel) or hasattr(obj, "model_dump"):
        if _xxhash_available:
            return xxhash.xxh64(obj.model_dump_json()).hexdigest()
        return hashlib.sha256(obj.model_dump_json().encode()).hexdigest()
    if isinstance(obj, dict):
        if _xxhash_available:
            return xxhash.xxh64(str(obj).encode()).hexdigest()
        return hashlib.sha256(str(obj).encode()).hexdigest()
    if _xxhash_available:
        return xxhash.xxh64(str(obj).encode()).hexdigest()
    return hashlib.sha256(str(obj).encode()).hexdigest()

class BaseSerializer(abc.ABC):
    """
    The Base Serializer Class
    """
    name: Optional[str] = None
    encoding: Optional[str] = None
    binary: Optional[bool] = False
    compressor: Optional['CompressionT'] = None
    previous_compressor: Optional['CompressionT'] = None

    def __init__(
        self,
        compression: Optional[str] = None,
        compression_level: Optional[int] = None,
        encoding: Optional[str] = None,
        raise_errors: bool = False,
        enable_deprecation_support: bool = True,
        **kwargs,
    ):
        """
        Initializes the serializer
        """
        if compression is not None or compression_level is not None:
            from ..compression import get_compression
            compression_kwargs = kwargs.pop("compression_kwargs", None)
            decompression_kwargs = kwargs.pop("decompression_kwargs", None)
            deprecated_compression = kwargs.pop("deprecated_compression", None)
            self.compressor = get_compression(
                compression, 
                compression_level = compression_level, 
                compression_kwargs = compression_kwargs, 
                decompression_kwargs = decompression_kwargs,
            )
            if deprecated_compression is not None and deprecated_compression != compression:
                self.previous_compressor = get_compression(deprecated_compression)
        if encoding is not None: self.encoding = encoding
        self.raise_errors = raise_errors
        self.enable_deprecation_support = enable_deprecation_support
        self._kwargs = kwargs

    @property
    def compression_enabled(self) -> bool:
        """
        Returns if compression is enabled
        """
        return self.compressor is not None
    
    @property
    def compression_level(self) -> Optional[int]:
        """
        Returns the compression level
        """
        return self.compressor.compression_level if self.compressor is not None else None

    def fetch_object_classname(self, obj: ObjectValue) -> str:
        """
        Fetches the object classname
        """
        return f"{obj.__class__.__module__}.{obj.__class__.__name__}"
    
    def create_hash(self, obj: ObjectValue) -> str:
        """
        Creates a hash for the object
        """
        return create_obj_hash(obj)
    
    def compress_value(self, value: Union[str, bytes], **kwargs) -> Union[str, bytes]:
        """
        Compresses the value
        """
        if self.compression_enabled:
            if isinstance(value, str): value = value.encode(self.encoding)
            return self.compressor.compress(value)
        return value
    
    def deprecated_decompress_value(self, value: Union[str, bytes], **kwargs) -> Optional[Union[str, bytes]]:
        """
        Attempts to decompress the value using the deprecated compressor
        """
        e = None
        attempt_msg = f"{self.name}"
        if self.previous_compressor is not None:
            try:
                return self.previous_compressor.decompress(value)
            except Exception as e:
                attempt_msg += f"-> {self.previous_compressor.name}"
        try:
            return zlib.decompress(value)
        except Exception as e:
            attempt_msg += " -> ZLib"
            logger.trace(f'[{attempt_msg}] Error in Decompression: {str(value)[:100]}', e)
            if self.raise_errors: raise e
            return None
        
    
    def decompress_value(self, value: Union[str, bytes], **kwargs) -> Union[str, bytes]:
        # sourcery skip: extract-duplicate-method
        """
        Decompresses the value
        """
        if not self.compression_enabled: return value
        try:
            value = self.compressor.decompress(value, **kwargs)
        except Exception as e:
            if self.enable_deprecation_support or self.previous_compressor is not None:
                value = self.deprecated_decompress_value(value, **kwargs)
        if value is not None and not self.binary: value = value.decode(self.encoding)
        return value

    def encode_value(self, value: ObjectValue, **kwargs) -> Union[str, bytes]:
        """
        Encodes the value
        """
        raise NotImplementedError
    
    def encode(self, value: ObjectValue, **kwargs) -> Union[str, bytes]:
        """
        Encodes the value
        """
        return self.compress_value(self.encode_value(value, **kwargs))
    
    async def aencode(self, value: ObjectValue, **kwargs) -> Union[str, bytes]:
        """
        Encodes the value asynchronously
        """
        return await ThreadPooler.run_async(self.encode, value, **kwargs)
    
    def decode_value(self, value: Union[str, bytes], **kwargs) -> ObjectValue:
        """
        Decodes the value
        """
        raise NotImplementedError
    
    def decode(self, value: Union[str, bytes], **kwargs) -> ObjectValue:
        """
        Decodes the value
        """
        return self.decode_value(self.decompress_value(value, **kwargs), **kwargs)
    
    async def adecode(self, value: Union[str, bytes], **kwargs) -> ObjectValue:
        """
        Decodes the value asynchronously
        """
        return await ThreadPooler.run_async(self.decode, value, **kwargs)
    
    def dumps(self, value: ObjectValue, **kwargs) -> Union[str, bytes]:
        # sourcery skip: class-extract-method
        """
        Dumps the value
        """
        try:
            return self.encode(value, **kwargs)
        except Exception as e:
            logger.trace(f'[{self.name}] Error in Encoding: {str(value)[:500]}', e)
            if self.raise_errors: raise e
            return None
        
    async def adumps(self, value: ObjectValue, **kwargs) -> Union[str, bytes]:
        """
        Dumps the value asynchronously
        """
        return await ThreadPooler.run_async(self.dumps, value, **kwargs)
    
    def loads(self, value: Union[str, bytes], **kwargs) -> ObjectValue:
        """
        Loads the value
        """
        try:
            return self.decode(value, **kwargs)
        except Exception as e:
            logger.trace(f'[{self.name}] Error in Decoding: {str(value)[:500]}', e)
            if self.raise_errors: raise e
            return None
        
    async def aloads(self, value: Union[str, bytes], **kwargs) -> ObjectValue:
        """
        Loads the value asynchronously
        """
        return await ThreadPooler.run_async(self.loads, value, **kwargs)
    
    

class BinaryBaseSerializer(BaseSerializer):

    binary: Optional[bool] = True




from __future__ import annotations

import os
import signal
import contextlib
import multiprocessing

from typing import Optional, List, TypeVar, Callable, Dict, Any, Union, TYPE_CHECKING
from lazyops.utils.lazy import lazy_import
from lazyops.imports._psutil import _psutil_available

if _psutil_available:
    import psutil

if TYPE_CHECKING:
    from lazyops.utils.logs import Logger
    from lazyops.types.models import BaseSettings
    from lazyops.libs.fastapi_utils.types.state import AppState

    with contextlib.suppress(ImportError):
        from aiokeydb.types.task_queue import TaskQueue
        from aiokeydb.types.worker import Worker

_base_worker_index: int = None

def get_base_worker_index() -> int:
    """
    Gets the base worker index
    """
    global _base_worker_index
    if _base_worker_index is None:
        from lazyops.utils.system import is_in_kubernetes, get_host_name
        if is_in_kubernetes() and get_host_name()[-1].isdigit():
            _base_worker_index = int(get_host_name()[-1])
        else:
            _base_worker_index = 0
    return _base_worker_index


SettingsT = TypeVar('SettingsT', bound='BaseSettings')

class GlobalContextMeta(type):
    """
    Global Context for FastAPI 
    """

    queues: Dict[str, Dict[str, 'TaskQueue']] = {}
    server_processes: List['multiprocessing.Process'] = []
    worker_processes: Dict[str, Dict[str, List['multiprocessing.Process']]] = {}

    on_close_funcs: List[Callable] = []

    settings_func: Optional[Union[Callable, str]] = None # type: ignore
    settings_config_func: Optional[Union[Callable, str]] = None # type: ignore
    get_worker_func: Optional[Union[Callable, str]] = None # type: ignore
    get_num_worker_func: Optional[Union[Callable, str]] = None # type: ignore
    get_worker_names_func: Optional[Union[Callable, str]] = None
    
    _settings: Optional[SettingsT] = None
    _state: Optional['AppState'] = None
    _logger: Optional['Logger'] = None

    def configure(
        cls,
        queue_types: Optional[List[str]] = None,
        settings_func: Optional[Union[Callable, str]] = None, # type: ignore
        settings_config_func: Optional[Union[Callable, str]] = None, # type: ignore
        get_worker_func: Optional[Union[Callable, str]] = None, # type: ignore
        get_num_worker_func: Optional[Union[Callable, str]] = None, # type: ignore
        get_worker_names_func: Optional[Union[Callable, str]] = None,
    ):
        """
        Configures the global context
        """
        if queue_types is not None:
            for kind in queue_types:
                cls.add_queue_type(kind)
        
        if settings_func is not None:
            if isinstance(settings_func, str):
                settings_func = lazy_import(settings_func)
            cls.settings_func = settings_func
        
        if settings_config_func is not None:
            cls.settings_config_func = settings_config_func
        
        if get_worker_func is not None:
            cls.get_worker_func = get_worker_func
        
        if get_num_worker_func is not None:
            cls.get_num_worker_func = get_num_worker_func
        
        if get_worker_names_func is not None:
            cls.get_worker_names_func = get_worker_names_func
            

    @property
    def settings(cls) -> SettingsT:
        """
        Returns the settings
        """
        if cls._settings is None:
            if cls.settings_func is None:
                raise RuntimeError("Settings not initialized. Set `GlobalContext.settings_func` to a function that returns the settings.")
            cls._settings = cls.settings_func()
        return cls._settings
    
    @property
    def state(cls) -> 'AppState':
        """
        Returns the state
        """
        if cls._state is None:
            from lazyops.libs.fastapi_utils.types.state import AppState
            cls._state = AppState()
            cls._state.bind_settings(cls.settings)
        return cls._state
    
    @property
    def is_leader_process(cls) -> bool:
        """
        Returns if this is the leader process
        """
        return cls.state.is_leader_process or multiprocessing.parent_process() is None
    
    @property
    def is_primary_server_process(cls) -> bool:
        """
        Returns if this is the primary server process
        """
        return cls.state.is_primary_server_process or multiprocessing.parent_process() is None
    

    def get_settings_func(cls, func: Union[Callable, str]) -> Callable:
        """
        Returns the settings func
        """
        if isinstance(func, str):
            func = getattr(cls.settings, func)
        return func
    

    @property
    def logger(cls) -> 'Logger':
        """
        Returns the logger
        """
        if cls._logger is None:
            from lazyops.utils.logs import logger as _logger
            cls._logger = _logger
        return cls._logger

    def get_child_process_ids(cls, name: Optional[str] = None, kind: Optional[str] = None, process_id: Optional[int] = None, first: Optional[bool] = True) -> List[int]:
        """
        Returns the child process ids
        """
        if not _psutil_available:
            cls.logger.warning("psutil not available. Cannot get child process ids")
            return []
        if not name and not process_id:
            raise ValueError("Must provide either name or process_id")
        if process_id:
            proc = psutil.Process(process_id)
            return [child.pid for child in proc.children(recursive=True)]
        if name in {'server', 'app'} and kind is None:
            procs = []
            for proc in cls.server_processes:
                if proc._closed: continue
                parent = psutil.Process(proc.pid)
                procs.extend(child.pid for child in parent.children(recursive=True))
                if first: break
            return procs
        if kind is None: kind = 'default'
        if kind not in cls.worker_processes:
            cls.logger.warning(f"No worker processes found for {kind}")
            return []
        if name not in cls.worker_processes[kind]:
            cls.logger.warning(f"No worker processes found for {kind}.{name}")
            return []
        procs = []
        for proc in cls.worker_processes[kind][name]:
            if proc._closed: continue
            parent = psutil.Process(proc.pid)
            procs.extend(child.pid for child in parent.children(recursive=True))
            if first: break
        return procs
    
    def add_queue_type(cls, kind: str):
        """
        Add a queue type
        """
        if kind not in cls.queues:
            setattr(cls, kind, {})
            cls.queues[kind] = getattr(cls, kind)
    
    def set_queue(cls, name: str, queue: 'TaskQueue', kind: Optional[str] = None):
        """
        Set a queue
        """
        if kind is None: kind = 'default'
        if kind not in cls.queues:
            cls.add_queue_type(kind)
        cls.queues[kind][name] = queue
    
    def add_worker_processes(cls, name: str, procs: List['multiprocessing.Process'], kind: Optional[str] = None):
        """
        Adds worker processes
        """
        if kind is None: kind = 'default'
        if cls.worker_processes.get(kind) is None: cls.worker_processes[kind] = {}
        if cls.worker_processes[kind].get(name) is None: cls.worker_processes[kind][name] = []
        cls.worker_processes[kind][name].extend(procs)

    def has_worker_processes(cls, name: str, kind: Optional[str] = None) -> bool:
        """
        Checks if there are processes
        """
        if kind is None: kind = 'default'
        if kind not in cls.worker_processes: cls.worker_processes[kind] = {}
        if name not in cls.worker_processes[kind]:
            cls.worker_processes[kind][name] = []        
        return len(cls.worker_processes[kind][name]) > 0
    
    def start_worker_processes(
        cls, 
        name: str, 
        kind: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None, 
        num_workers: Optional[int] = 1, 
        verbose: Optional[bool] = True, 
        base_index: Optional[Union[int, str]] = 'auto',

        spawn_worker_func: Optional[Union[Callable, str]] = None,
        num_worker_func: Optional[Union[Callable, str]] = None,
        get_worker_func: Optional[Union[Callable, str]] = None,
        settings_config_func: Optional[Union[Callable, str]] = None,

        **kwargs
    ):  # sourcery skip: low-code-quality
        """
        Starts the worker processes
        """
        procs = []
        if kind is None: kind = 'default'
        if verbose: cls.logger.info(f"[{kind.capitalize()}] Spawning Worker: {name} ({num_workers})")
        context = multiprocessing.get_context('spawn')
        if base_index == 'auto':
            base_index = get_base_worker_index()
        if num_worker_func is None: num_worker_func = cls.get_settings_func(cls.get_num_worker_func)
        if get_worker_func is None: get_worker_func = cls.get_settings_func(cls.get_worker_func)
        if settings_config_func is None: 
            settings_config_func = cls.get_settings_func(cls.settings_config_func)
        if spawn_worker_func is None: 
            from .workers import spawn_new_worker
            spawn_worker_func = spawn_new_worker
        
        if num_worker_func:
            num_workers = num_worker_func(name = name, num_workers = num_workers, kind = kind)
        if not kwargs: kwargs = {}
        
        kwargs['kind'] = kind
        kwargs['config'] = config
        kwargs['verbose'] = verbose

        kwargs['settings_config_func'] = settings_config_func
        kwargs['get_worker_func'] = get_worker_func
        
        for n in range(num_workers):
            is_primary_worker = n == 0
            worker_index = (base_index * num_workers) + n
            kwargs['is_primary_worker'] = is_primary_worker
            kwargs['index'] = worker_index
            p = context.Process(target = spawn_worker_func, args = (name,), kwargs = kwargs)
            p.start()
            if verbose: 
                log_name = f'{name}-{worker_index}'
                cls.logger.info(f"- |g|[{kind.capitalize()}]|e| Started: [ {n + 1}/{num_workers} ] `|g|{log_name:20s}|e|` (Process ID: {p.pid})", colored = True)
            procs.append(p)
        cls.add_worker_processes(kind = kind, name = name, procs = procs)
        return procs
    
    def start_all_workers(
        cls,
        worker_names: Optional[List[str]] = None,
        kind: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        num_workers: Optional[int] = 1,
        base_index: Optional[Union[int, str]] = 'auto',


        enabled_workers: Optional[List[str]] = None,
        disabled_workers: Optional[List[str]] = None,
        verbose: Optional[bool] = True,

        spawn_worker_func: Optional[Union[Callable, str]] = None,
        num_worker_func: Optional[Union[Callable, str]] = None,
        get_worker_func: Optional[Union[Callable, str]] = None,
        settings_config_func: Optional[Union[Callable, str]] = None,
        get_worker_names_func: Optional[Union[Callable, str]] = None,
        **kwargs
    ):  # sourcery skip: low-code-quality
        """
        Starts all the workers
        """
        procs = []
        if kind is None: kind = 'default'
        if not worker_names:
            if get_worker_names_func is None:
                get_worker_names_func = cls.get_settings_func(cls.get_worker_names_func)
            worker_names = get_worker_names_func(
                kind = kind,
                enabled_workers = enabled_workers,
                disabled_workers = disabled_workers,
            )
        if verbose: cls.logger.info(f"[|g|{kind.capitalize()}|e|] Starting {len(worker_names)} Workers: {worker_names}", colored = True)
        context = multiprocessing.get_context('spawn')
        if base_index == 'auto': base_index = get_base_worker_index()
        if num_worker_func is None: num_worker_func = cls.get_settings_func(cls.get_num_worker_func)
        if get_worker_func is None: get_worker_func = cls.get_settings_func(cls.get_worker_func)
        if settings_config_func is None: 
            settings_config_func = cls.get_settings_func(cls.settings_config_func)
        if spawn_worker_func is None: 
            from .workers import spawn_new_worker
            spawn_worker_func = spawn_new_worker
        
        if not kwargs: kwargs = {}
        
        kwargs['kind'] = kind
        kwargs['config'] = config
        kwargs['verbose'] = verbose

        kwargs['settings_config_func'] = settings_config_func
        kwargs['get_worker_func'] = get_worker_func
        
        for name in worker_names:
            if num_worker_func: num_workers = num_worker_func(name = name, num_workers = num_workers, kind = kind)
            
            for n in range(num_workers):
                is_primary_worker = n == 0
                worker_index = (base_index * num_workers) + n
                kwargs['is_primary_worker'] = is_primary_worker
                kwargs['index'] = worker_index
                p = context.Process(target = spawn_new_worker, args = (name,), kwargs = kwargs)
                p.start()
                if verbose: 
                    log_name = f'`|g|{name}-{worker_index}|e|`'
                    cls.logger.info(f"- [|g|{kind.capitalize():7s}|e|] Started: [ {n + 1}/{num_workers} ] {log_name:20s} (Process ID: {p.pid})", colored = True)
                if is_primary_worker:
                    cls.state.add_leader_process_id(p.pid, kind)
                
                procs.append(p)
            cls.add_worker_processes(kind = kind, name = name, procs = procs)
        return procs

    def stop_worker_processes(cls,  name: str, verbose: bool = True, timeout: float = 5.0, kind: Optional[str] = None):
        # sourcery skip: low-code-quality
        """
        Stops the worker processes
        """
        if kind is None: kind = 'default'
        if cls.worker_processes.get(kind) is None or cls.worker_processes[kind].get(name) is None: 
            if verbose: cls.logger.warning(f"[{kind.capitalize()}] No worker processes found for {name}")
            return
        curr_proc, n_procs = 0, len(cls.worker_processes[kind][name])

        while cls.worker_processes[kind][name]:
            proc = cls.worker_processes[kind][name].pop()
            if proc._closed: continue
            log_name = f'`|g|{name}-{curr_proc}|e|`'
            if verbose: 
                cls.logger.info(f"- [|y|{kind.capitalize():7s}|e|] Stopping: [ {curr_proc + 1}/{n_procs} ] {log_name:20s} (Process ID: {proc.pid})", colored = True)
            proc.join(timeout)
            proc.terminate()
            try:
                proc.close()
            except Exception as e:
                if verbose: cls.logger.info(f"|y|[{kind.capitalize():7s}]|e| Failed Stop: [ {curr_proc + 1}/{n_procs} ] {log_name:20s} (Error: |r|{e}|e|)", colored = True)
                try:
                    signal.pthread_kill(proc.ident, signal.SIGKILL)
                    proc.join(timeout)
                    proc.terminate()
                except Exception as e:
                    if verbose: cls.logger.info(f"|r|[{kind.capitalize():7s}]|e| Failed Kill: [ {curr_proc + 1}/{n_procs} ] {log_name:20s} (Error: |r|{e}|e|)", colored = True)
                with contextlib.suppress(Exception):
                    proc.kill()
                    proc.close()
            curr_proc += 1

    async def astop_worker_processes(cls, name: str, verbose: bool = True, timeout: float = 5.0, kind: Optional[str] = None):
        """
        Stops the worker processes
        """
        if kind is None: kind = 'default'
        if cls.worker_processes.get(kind) is None or cls.worker_processes[kind].get(name) is None: 
            if verbose: cls.logger.warning(f"[{kind.capitalize()}] No worker processes found for {name}")
            return
        from .workers import terminate_worker
        await terminate_worker(name = name, kind = kind)
        cls.stop_worker_processes(name = name, verbose = verbose, timeout = timeout, kind = kind)
    
    def start_server_process(cls, cmd: str, verbose: bool = True):
        """
        Starts the server process
        """
        if verbose: cls.logger.info(f"Starting Server Process: {cmd}")
        context = multiprocessing.get_context('spawn')
        p = context.Process(target = os.system, args = (cmd,))
        p.start()
        cls.state.add_leader_process_id(p.pid, 'server')
        cls.server_processes.append(p)
        return p
    
    def add_server_process(cls, name: str, cmd: str, verbose: bool = True):
        """
        Adds the server process
        """
        if verbose: cls.logger.info(f"Adding Server Process: {cmd}", prefix = name)
        context = multiprocessing.get_context('spawn')
        p = context.Process(target = os.system, args = (cmd,))
        p.start()
        cls.server_processes.append(p)
        return p

    def stop_server_processes(cls, verbose: bool = True, timeout: float = 5.0):
        """
        Stops the server processes
        """
        curr_proc, n_procs = 0, len(cls.server_processes)
        kind = 'Server'
        while cls.server_processes:
            proc = cls.server_processes.pop()
            if proc._closed: continue
            log_name = f'`|g|app-{curr_proc}|e|`'
            if verbose: 
                cls.logger.info(f"- [|y|{kind:7s}|e|] Stopping: [ {curr_proc + 1}/{n_procs} ] {log_name:20s} (Process ID: {proc.pid})", colored = True)
            proc.join(timeout)
            proc.terminate()
            try:
                proc.close()
            except Exception as e:
                if verbose: cls.logger.info(f"|y|[{kind:7s}]|e| Failed Stop: [ {curr_proc + 1}/{n_procs} ] {log_name:20s} (Error: |r|{e}|e|)", colored = True)
                try:
                    signal.pthread_kill(proc.ident, signal.SIGKILL)
                    proc.join(timeout)
                    proc.terminate()
                except Exception as e:
                    if verbose: cls.logger.info(f"|r|[{kind:7s}]|e| Failed Kill: [ {curr_proc + 1}/{n_procs} ] {log_name:20s} (Error: |r|{e}|e|)", colored = True)
                with contextlib.suppress(Exception):
                    proc.kill()
                    proc.close()
            curr_proc += 1
        # if verbose: cls.logger.info(f"Stopped all {curr_proc} server processes")
    
    def end_all_processes(cls, verbose: bool = True, timeout: float = 5.0):
        """
        Terminates all processes
        """
        for kind, names in cls.worker_processes.items():
            for name in names:
                cls.stop_worker_processes(name = name, verbose = verbose, timeout = timeout, kind = kind)
        cls.stop_server_processes(verbose = verbose, timeout = timeout)
        # cls.logger.info("Terminated all processes")

    def register_on_close(cls, func: Callable, *args, **kwargs):
        """
        Registers a function to be called on close
        """
        import functools
        _func = functools.partial(func, *args, **kwargs)
        cls.on_close_funcs.append(_func)
        cls.logger.info(f"Registered function {func.__name__} to be called on close")


    async def aclose_processes(cls, verbose: bool = True, timeout: float = 5.0):
        """
        Handles closing all processes
        """
        import anyio
        for func in cls.on_close_funcs:
            with anyio.fail_after(timeout):
                try:
                    await func()
                    if verbose: cls.logger.info(f"Called function {func.__name__} on close")
                except TimeoutError as e:
                    cls.logger.error(f"Timeout calling function {func.__name__} on close: {e}")
                except Exception as e:
                    cls.logger.trace(f"Error calling function {func.__name__} on close", error = e)
    
    def start_debug_mode(cls):
        """
        Enters Debug Mode
        """
        cls.logger.warning('Entering Debug Mode. Sleeping Forever')
        import time
        import sys
        while True:
            try:
                time.sleep(900)
            except Exception as e:
                cls.logger.error(e)
                break
        sys.exit(0)



class GlobalContext(metaclass=GlobalContextMeta):
    """
    Global Context for Workers and Builders
    """


class GracefulKiller:
    kill_now = False
    signals = {
        signal.SIGINT: 'SIGINT',
        signal.SIGTERM: 'SIGTERM',
        signal.SIGKILL: 'SIGKILL',
        signal.SIGQUIT: 'SIGQUIT',
        signal.SIGSTOP: 'SIGSTOP',
        signal.SIGABRT: 'SIGABRT',
    }

    def __init__(
        self, 
        enabled: Optional[List[signal.Signals]] = [signal.SIGINT, signal.SIGTERM],
        logger: Optional['Logger'] = None
    ):  # sourcery skip: default-mutable-arg
        
        self.enabled = enabled
        if logger is None:
            from lazyops.utils.logs import logger as _logger
            logger = _logger
        self.logger = logger
        for sig in self.enabled:
            signal.signal(sig, self.exit_gracefully)
        self.pid = os.getpid()

    def exit_gracefully(self, signum, frame):
        self.logger.warning(f"[{self.pid}] Received {self.signals[signum]} signal. Exiting gracefully")
        self.kill_now = True


from lazyops import get_logger, timer
from lazyops.apis import LazyAPI

## Eleuther.ai API
## Source: https://github.com/kingoflolz/mesh-transformer-jax/#gpt-j-6b

eai_config = {
    'url': 'https://api.eleuther.ai',
    'token': None,
    'default_params': {'top_p': 0.9, 'temp': 1},
    'params_key': None,
    'data_key': 'context',
    'default_fetch': 'complete',
    'default_async': 'complete_async',
    'route_config': {
        'complete': {
            'method': 'POST',
            'path': '/complete',
            'params': {'top_p': 0.9, 'temp': 1},
            'params_key': None,
            'data_key': 'context',
            'is_async': False,
            'prefix_payload': False,
            'decode_json': True,
        },
        'complete_async': {
            'method': 'POST',
            'path': '/complete',
            'params': {'top_p': 0.9, 'temp': 1},
            'params_key': None,
            'data_key': 'context',
            'is_async': True,
            'prefix_payload': False,
            'decode_json': True,
        },
    }
}

# TextSynth API
# Source: https://bellard.org/textsynth/


textsynth_config = {
    'url': 'https://bellard.org/textsynth/api/v1/engines/gptj_6B',
    'token': None,
    'default_params': {'top_p': 0.9, 'temperature': 1, 'seed': 0, 'stream': False, 'top_k': 40},
    'params_key': None,
    'data_key': 'prompt',
    'default_fetch': 'complete',
    'default_async': 'complete_async',
    'route_config': {
        'complete': {
            'method': 'POST',
            'path': '/completions',
            'params': {'top_p': 0.9, 'temperature': 1, 'seed': 0, 'stream': False, 'top_k': 40},
            'params_key': None,
            'data_key': 'prompt',
            'is_async': False,
            'prefix_payload': False,
            'decode_json': True,
        },
        'complete_async': {
            'method': 'POST',
            'path': '/completions',
            'params': {'top_p': 0.9, 'temperature': 1, 'seed': 0, 'stream': False, 'top_k': 40},
            'params_key': None,
            'data_key': 'prompt',
            'is_async': True,
            'prefix_payload': False,
            'decode_json': True,
        },
    }
}

logger = get_logger('LazyEXP', 'GPTJ')

class GPTJAPI:
    def __init__(self, fallback='eai'):
        self.eai = LazyAPI.build(eai_config)
        self._eai_config =  {'top_p': 0.9, 'temp': 1}
        self.ts = LazyAPI.build(textsynth_config)
        self._ts_config = {'top_p': 0.9, 'temperature': 1, 'seed': 0, 'stream': False, 'top_k': 40}
        self._fallback = fallback
    
    @property
    def primary(self):
        if self._fallback == 'eai':
            return self.ts
        return self.eai
    
    @property
    def fallback(self):
        if self._fallback == 'eai':
            return self.eai
        return self.ts
    
    def get_base_params(self, is_fallback=False):
        if is_fallback:
            if self._fallback == 'eai':
                return self._eai_config.copy()
            return self._ts_config.copy()
        if self._fallback == 'eai':
            return self._ts_config.copy()
        return self._eai_config.copy()

    def build_params(self, text, is_fallback=False, **config):
        p = self.get_base_params(is_fallback)
        if config:
            for k,v in config.items():
                if k in p:
                    config[k] = v
        return p

    def predict(self, text, **config):
        p = self.build_params(text, **config)
        try:
            res = self.primary.complete(data=text, **p)
            #print(res)
            return res.get('completion', res.get('text','')), res
        
        except Exception as e:
            logger.error(f'Error Calling API: {str(e)}')
            p = self.build_params(text, is_fallback=True, **config)
            try:
                res = self.fallback.complete(data=text, **p)
                return res.get('text', res.get('completion','')), res
            except Exception as e:
                logger.error(f'Error Calling API: {str(e)}')
                return None, None
    
    async def as_predict(self, text, **config):
        p = self.build_params(text, **config)
        try:
            res = await self.primary.complete_async(**p)
            return res.get('completion', res.get('text','')), res
        
        except Exception as e:
            logger.error(f'Error Calling API: {str(e)}')
            p = self.build_params(text, is_fallback=True, **config)
            try:
                res = await self.fallback.complete_async(**p)
                return res.get('text', res.get('completion','')), res
            except Exception as e:
                logger.error(f'Error Calling API: {str(e)}')
                return None, None
    
    
                

if __name__ == '__main__':
    test_texts = [
        'In a shocking finding, scientist discovered a herd of unicorns living in a remote, previously unexplored valley, in the Andes Mountains. Even more surprising to the researchers was the fact that the unicorns spoke perfect English.',
        'Game of Thrones is',
        'For todays homework assignment, please describe the reasons for the US Civil War.'
    ]
    api = GPTJAPI()
    for txt in test_texts:
        res, _ = api.predict(text=txt)
        logger.info(res)
    


    







"""
Import Handler for aiohttpx
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import aiohttpx
    _aiohttpx_available = True
except ImportError:
    aiohttpx = object
    _aiohttpx_available = False

def resolve_aiohttpx(
    required: bool = False,
):
    """
    Ensures that `aiohttpx` is available
    """
    global aiohttpx, _aiohttpx_available
    if not _aiohttpx_available:
        resolve_missing('aiohttpx', required = required)
        import aiohttpx
        _aiohttpx_available = True
        globals()['aiohttpx'] = aiohttpx


def require_aiohttpx(
    required: bool = False,
):
    """
    Wrapper for `resolve_aiohttpx` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_aiohttpx, 
            func = func, 
            required = required
        )
    return decorator


"""
Retrieval Utilities
"""

import os
import re
import base64
import requests
import json
import functools
import contextlib
from pathlib import Path
from typing import Optional, Union, Tuple, List, Dict, Any, TYPE_CHECKING
from lazyops.utils.logs import logger
from lazyops.types import BaseModel, lazyproperty, Literal
from pydantic.types import ByteSize


with contextlib.suppress(ImportError):
    from tqdm.auto import tqdm as base_tqdm

user_home = Path(os.path.expanduser("~"))



"""
HuggingFace Utilities
"""


def get_huggingface_token(
    raise_error: Optional[bool] = False,
) -> Optional[str]:
    """
    Gets the HuggingFace API token
    """
    # check env first
    for key in {'HF_TOKEN', 'HUGGINGFACE_TOKEN', 'HUGGINGFACE_API_TOKEN'}:
        if key in os.environ:
            return os.environ[key]
    
    token_path = user_home.joinpath(".huggingface/token")
    if token_path.exists():
        with open(token_path, 'r') as f:
            return f.read().strip()
    
    if raise_error:
        raise ValueError("HuggingFace API token not found. Please run `huggingface-cli login` to login.")
    return None


def sanitize_branch_name(branch_name: str):
    pattern = re.compile(r"^[a-zA-Z0-9._-]+$")
    if pattern.match(branch_name):
        return branch_name
    raise ValueError("Invalid branch name. Only alphanumeric characters, period, underscore and dash are allowed.")

re_patterns = {
    'adapter': re.compile(r"adapter.*\.bin"),
    'pytorch': re.compile(r".*\.pt"),
    'pytorch_model': re.compile(r"(pytorch|adapter)_model.*\.bin"),
    'safetensors': re.compile(r"model.*\.safetensors"),
    'ggml': re.compile(r".*\.ggml.*\.bin"),
    'tensorflow': re.compile(r".*\.(tf|h5)"),
    'tokenizer': re.compile(r"(tokenizer|special_tokens_map|spiece).*\.(json|model)"),
    'text': re.compile(r".*\.(txt|json|py|md)"),
    'config': re.compile(r".*\.json"),
    'flax': re.compile(r"flax*"),
}
reg_patterns = {
    'adapter': ["adapter*.bin"],
    'pytorch': ["*.pt", 'pytorch*.bin'],
    'safetensors': ["*.safetensors"],
    'ggml': ["*ggml*.bin"],
    'tensorflow': ["*.h5", "*.tf"],
    'tokenizer': ["tokenizer*.json", "special_tokens_map*.json", "spiece*.model"],
    'text': ["*.txt", "*.json", "*.py", "*.md"],
    'config': ["*.json"],
    'flax': ["flax*."],
}

# re_compile_patterns = {
#     key: re.compile(value) for key, value in re_patterns.items()
# }

class HFLink(BaseModel):
    url: str
    filename: str
    size: ByteSize

    @lazyproperty
    def is_pytorch_model(self):
        return bool(re_patterns['pytorch_model'].match(self.filename))
    
    @lazyproperty
    def is_safetensors(self):
        return bool(re_patterns['safetensors'].match(self.filename))
    
    @lazyproperty
    def is_pytorch(self):
        return bool(re_patterns['pytorch'].match(self.filename))
    
    @lazyproperty
    def is_tensorflow(self):
        return bool(re_patterns['tensorflow'].match(self.filename))
    
    @lazyproperty
    def is_tokenizer(self):
        return bool(re_patterns['tokenizer'].match(self.filename))
    
    @lazyproperty
    def is_text(self):
        return bool(re_patterns['text'].match(self.filename))
    
    @lazyproperty
    def is_lora(self):
        return self.filename.endswith(('adapter_config.json', 'adapter_model.bin'))
    
    @lazyproperty
    def is_config(self):
        return bool(re_patterns['config'].match(self.filename))

class HFLinks(BaseModel):
    repo_id: str
    repo_type: str
    revision: str

    links: List[HFLink] = []

    @lazyproperty
    def is_lora(self):
        return any(link.is_lora for link in self.links)

    @lazyproperty
    def lora_links(self):
        return [link for link in self.links if link.is_lora]

    @lazyproperty
    def pytorch_links(self):
        return [link for link in self.links if link.is_pytorch]
    
    @lazyproperty
    def pytorch_model_links(self):
        return [link for link in self.links if link.is_pytorch_model]
    
    @lazyproperty
    def safetensors_links(self):
        return [link for link in self.links if link.is_safetensors]
    
    @lazyproperty
    def tensorflow_links(self):
        return [link for link in self.links if link.is_tensorflow]
    
    @lazyproperty
    def config_links(self):
        return [link for link in self.links if link.is_config and not link.is_tokenizer]

    @lazyproperty
    def tokenizer_links(self):
        return [link for link in self.links if link.is_tokenizer]
    
    @lazyproperty
    def text_links(self):
        return [link for link in self.links if link.is_text and not link.is_tokenizer and not link.is_config]
    
    @lazyproperty
    def classifications(self) -> List[str]:
        res = []
        if self.lora_links:
            res.append('lora')
        if self.pytorch_model_links or self.pytorch_links:
            res.append('pytorch')
        if self.tensorflow_links:
            res.append('tensorflow')
        if self.safetensors_links:
            res.append('safetensors')
        return res


    def __len__(self):
        return len(self.links)
    
    def __getitem__(self, idx: Union[str, int]) -> Union[List[HFLink], HFLink]:
        """
        Returns the link at the given index or by the type
        """
        if isinstance(idx, int):
            return self.links[idx]
        if hasattr(self, f"{idx}_links"):
            return getattr(self, f"{idx}_links")
        # assume its a filename
        for link in self.links:
            if link.filename == idx:
                return link
        raise KeyError(f"Link with filename `{idx}` not found.")

@functools.lru_cache(maxsize=128)
def get_download_links_from_huggingface(
    repo_id: str,
    revision: Optional[str] = None,
    repo_type: Optional[str] = None,
    auth_token: Optional[Union[str, bool]] = True,
) -> HFLinks:
    """
    Retrieves the download links for a given `repo_id`.

    Args:
        repo_id (`str`):
            The ID of the repo on huggingface.co.
        revision (`str`, *optional*):
            The specific model version to use. Will default to `"main"` if it's not provided and no `commit_hash` is
            provided either.
        repo_type (`str`, *optional*):
            The type of the repository. Will default to `"models"`.
        auth_token (Union[`str`, `bool`, *optional*):
            The HuggingFace API token to use. 
            If `False`, will not use any token. If `None`, will try to get the token.
            If `True`, will try to get the token if found.
            If a string, will use that as the token.
    
    """
    if revision is None: revision = "main"
    if repo_type is None: repo_type = "models"

    repo_links = HFLinks(repo_id=repo_id, repo_type=repo_type, revision=revision)
    base = "https://huggingface.co"
    page = f"/api/{repo_type}/{repo_id}/tree/{revision}"
    cursor = b""

    headers = {}
    if auth_token:
        if isinstance(auth_token, str):
            headers['authorization'] = f'Bearer {auth_token}'
        elif auth_token is True:
            headers['authorization'] = f'Bearer {get_huggingface_token(False)}'

    while True:
        content = requests.get(
            f"{base}{page}",
            headers = headers,
            params = {'cursor': cursor.decode()} if cursor else None
        ).content
        data = json.loads(content)
        if len(data) == 0: break
        for i in range(len(data)):
            filename = data[i]['path']
            repo_links.links.append(
                HFLink(
                    url = f"https://huggingface.co/{repo_id}/resolve/{revision}/{filename}",
                    filename = filename,
                    size = data[i]['size'],
                )
            )
        
        cursor = base64.b64encode(f'{{"file_name":"{data[-1]["path"]}"}}'.encode()) + b':50'
        cursor = base64.b64encode(cursor)
        cursor = cursor.replace(b'=', b'%3D')

    return repo_links

"""
Forked from huggingface hub to ensure multithreading
"""

def snapshot_download(
    repo_id: str,
    *,
    revision: Optional[str] = None,
    repo_type: Optional[str] = None,
    cache_dir: Union[str, Path, None] = None,
    local_dir: Union[str, Path, None] = None,
    local_dir_use_symlinks: Union[bool, Literal["auto"]] = "auto",
    library_name: Optional[str] = None,
    library_version: Optional[str] = None,
    user_agent: Optional[Union[Dict, str]] = None,
    proxies: Optional[Dict] = None,
    etag_timeout: float = 10,
    resume_download: bool = False,
    force_download: bool = False,
    token: Optional[Union[bool, str]] = None,
    local_files_only: bool = False,
    allow_patterns: Optional[Union[List[str], str]] = None,
    ignore_patterns: Optional[Union[List[str], str]] = None,
    max_workers: int = 8,
    tqdm_class: Optional['base_tqdm'] = None,
) -> str:
    """Download repo files.

    Download a whole snapshot of a repo's files at the specified revision. This is useful when you want all files from
    a repo, because you don't know which ones you will need a priori. All files are nested inside a folder in order
    to keep their actual filename relative to that folder. You can also filter which files to download using
    `allow_patterns` and `ignore_patterns`.

    If `local_dir` is provided, the file structure from the repo will be replicated in this location. You can configure
    how you want to move those files:
      - If `local_dir_use_symlinks="auto"` (default), files are downloaded and stored in the cache directory as blob
        files. Small files (<5MB) are duplicated in `local_dir` while a symlink is created for bigger files. The goal
        is to be able to manually edit and save small files without corrupting the cache while saving disk space for
        binary files. The 5MB threshold can be configured with the `HF_HUB_LOCAL_DIR_AUTO_SYMLINK_THRESHOLD`
        environment variable.
      - If `local_dir_use_symlinks=True`, files are downloaded, stored in the cache directory and symlinked in `local_dir`.
        This is optimal in term of disk usage but files must not be manually edited.
      - If `local_dir_use_symlinks=False` and the blob files exist in the cache directory, they are duplicated in the
        local dir. This means disk usage is not optimized.
      - Finally, if `local_dir_use_symlinks=False` and the blob files do not exist in the cache directory, then the
        files are downloaded and directly placed under `local_dir`. This means if you need to download them again later,
        they will be re-downloaded entirely.

    An alternative would be to clone the repo but this requires git and git-lfs to be installed and properly
    configured. It is also not possible to filter which files to download when cloning a repository using git.

    Args:
        repo_id (`str`):
            A user or an organization name and a repo name separated by a `/`.
        revision (`str`, *optional*):
            An optional Git revision id which can be a branch name, a tag, or a
            commit hash.
        repo_type (`str`, *optional*):
            Set to `"dataset"` or `"space"` if downloading from a dataset or space,
            `None` or `"model"` if downloading from a model. Default is `None`.
        cache_dir (`str`, `Path`, *optional*):
            Path to the folder where cached files are stored.
        local_dir (`str` or `Path`, *optional*:
            If provided, the downloaded files will be placed under this directory, either as symlinks (default) or
            regular files (see description for more details).
        local_dir_use_symlinks (`"auto"` or `bool`, defaults to `"auto"`):
            To be used with `local_dir`. If set to "auto", the cache directory will be used and the file will be either
            duplicated or symlinked to the local directory depending on its size. It set to `True`, a symlink will be
            created, no matter the file size. If set to `False`, the file will either be duplicated from cache (if
            already exists) or downloaded from the Hub and not cached. See description for more details.
        library_name (`str`, *optional*):
            The name of the library to which the object corresponds.
        library_version (`str`, *optional*):
            The version of the library.
        user_agent (`str`, `dict`, *optional*):
            The user-agent info in the form of a dictionary or a string.
        proxies (`dict`, *optional*):
            Dictionary mapping protocol to the URL of the proxy passed to
            `requests.request`.
        etag_timeout (`float`, *optional*, defaults to `10`):
            When fetching ETag, how many seconds to wait for the server to send
            data before giving up which is passed to `requests.request`.
        resume_download (`bool`, *optional*, defaults to `False):
            If `True`, resume a previously interrupted download.
        force_download (`bool`, *optional*, defaults to `False`):
            Whether the file should be downloaded even if it already exists in the local cache.
        token (`str`, `bool`, *optional*):
            A token to be used for the download.
                - If `True`, the token is read from the HuggingFace config
                  folder.
                - If a string, it's used as the authentication token.
        local_files_only (`bool`, *optional*, defaults to `False`):
            If `True`, avoid downloading the file and return the path to the
            local cached file if it exists.
        allow_patterns (`List[str]` or `str`, *optional*):
            If provided, only files matching at least one pattern are downloaded.
        ignore_patterns (`List[str]` or `str`, *optional*):
            If provided, files matching any of the patterns are not downloaded.
        max_workers (`int`, *optional*):
            Number of concurrent threads to download files (1 thread = 1 file download).
            Defaults to 8.
        tqdm_class (`tqdm`, *optional*):
            If provided, overwrites the default behavior for the progress bar. Passed
            argument must inherit from `tqdm.auto.tqdm` or at least mimic its behavior.
            Note that the `tqdm_class` is not passed to each individual download.
            Defaults to the custom HF progress bar that can be disabled by setting
            `HF_HUB_DISABLE_PROGRESS_BARS` environment variable.

    Returns:
        Local folder path (string) of repo snapshot

    <Tip>

    Raises the following errors:

    - [`EnvironmentError`](https://docs.python.org/3/library/exceptions.html#EnvironmentError)
      if `token=True` and the token cannot be found.
    - [`OSError`](https://docs.python.org/3/library/exceptions.html#OSError) if
      ETag cannot be determined.
    - [`ValueError`](https://docs.python.org/3/library/exceptions.html#ValueError)
      if some parameter value is invalid

    </Tip>
    """
    from tqdm.contrib.concurrent import thread_map
    
    from huggingface_hub.constants import (
        DEFAULT_REVISION,
        HF_HUB_ENABLE_HF_TRANSFER,
        HUGGINGFACE_HUB_CACHE,
        REPO_TYPES,
        hf_cache_home,
    )
    from huggingface_hub.file_download import (
        REGEX_COMMIT_HASH, 
        hf_hub_download, 
        repo_folder_name
    )
    from huggingface_hub.hf_api import HfApi
    from huggingface_hub.utils import filter_repo_objects
    from huggingface_hub.utils import tqdm as hf_tqdm
    # from transformers.file_utils

    if repo_type is None: repo_type = "model"
    if repo_type not in REPO_TYPES:
        raise ValueError(f"Invalid repo type: {repo_type}. Accepted repo types are: {str(REPO_TYPES)}")

    if cache_dir is None: cache_dir = os.path.join(hf_cache_home, f"{repo_type}s")
        # cache_dir = HUGGINGFACE_HUB_CACHE
    if revision is None: revision = DEFAULT_REVISION
    if isinstance(cache_dir, Path): cache_dir = str(cache_dir)
    storage_folder = os.path.join(cache_dir, repo_folder_name(repo_id=repo_id, repo_type=repo_type))
    # if we have no internet connection we will look for an
    # appropriate folder in the cache
    # If the specified revision is a commit hash, look inside "snapshots".
    # If the specified revision is a branch or tag, look inside "refs".
    if local_files_only:
        if REGEX_COMMIT_HASH.match(revision):
            commit_hash = revision
        else:
            # retrieve commit_hash from file
            ref_path = os.path.join(storage_folder, "refs", revision)
            commit_hash = Path(ref_path).read_text()
        snapshot_folder = os.path.join(storage_folder, "snapshots", commit_hash)

        if os.path.exists(snapshot_folder):
            return snapshot_folder

        raise ValueError(
            "Cannot find an appropriate cached snapshot folder for the specified"
            " revision on the local disk and outgoing traffic has been disabled. To"
            " enable repo look-ups and downloads online, set 'local_files_only' to"
            " False."
        )

    # if we have internet connection we retrieve the correct folder name from the huggingface api
    _api = HfApi()
    repo_info = _api.repo_info(
        repo_id=repo_id,
        repo_type=repo_type,
        revision=revision,
        token=token,
    )
    assert repo_info.sha is not None, "Repo info returned from server must have a revision sha."
    filtered_repo_files = list(
        filter_repo_objects(
            items=[f.rfilename for f in repo_info.siblings],
            allow_patterns=allow_patterns,
            ignore_patterns=ignore_patterns,
        )
    )
    commit_hash = repo_info.sha
    snapshot_folder = os.path.join(storage_folder, "snapshots", commit_hash)
    # if passed revision is not identical to commit_hash
    # then revision has to be a branch name or tag name.
    # In that case store a ref.
    if revision != commit_hash:
        ref_path = os.path.join(storage_folder, "refs", revision)
        os.makedirs(os.path.dirname(ref_path), exist_ok=True)
        with open(ref_path, "w") as f:
            f.write(commit_hash)

    # we pass the commit_hash to hf_hub_download
    # so no network call happens if we already
    # have the file locally.
    def _inner_hf_hub_download(repo_file: str):
        return hf_hub_download(
            repo_id,
            filename=repo_file,
            repo_type=repo_type,
            revision=commit_hash,
            cache_dir=cache_dir,
            local_dir=local_dir,
            local_dir_use_symlinks=local_dir_use_symlinks,
            library_name=library_name,
            library_version=library_version,
            user_agent=user_agent,
            proxies=proxies,
            etag_timeout=etag_timeout,
            resume_download=resume_download,
            force_download=force_download,
            token=token,
        )

    if HF_HUB_ENABLE_HF_TRANSFER:
        # when using hf_transfer we don't want extra parallelism
        # from the one hf_transfer provides
        for file in filtered_repo_files:
            _inner_hf_hub_download(file)
    else:
        thread_map(
            _inner_hf_hub_download,
            filtered_repo_files,
            desc=f"Fetching {len(filtered_repo_files)} files",
            max_workers=max_workers,
            # User can use its own tqdm class or the default one from `huggingface_hub.utils`
            tqdm_class=tqdm_class or hf_tqdm,
        )

    if local_dir is not None:
        return str(os.path.realpath(local_dir))
    return snapshot_folder

def download_pretrained(
    repo_id: str,
    revision: Optional[str] = None,
    repo_type: Optional[str] = None,
    auth_token: Optional[Union[str, bool]] = True,
    classifications: Optional[List[str]] = None,
    cache_dir: Optional[Union[str, Path]] = None,
    allow_patterns: Optional[Union[List[str], str]] = None,
    **kwargs,
):
    """
    Handles the downloading of pretrained models from HuggingFace.
    """
    if classifications is None: 
        classifications = ['pytorch', 'tokenizer', 'config']
    if allow_patterns is None:
        allow_patterns = []
    if isinstance(allow_patterns, str):
        allow_patterns = [allow_patterns]
    for classification in classifications:
        allow_patterns.extend(reg_patterns[classification])
    if auth_token and auth_token is True:
        auth_token = get_huggingface_token(False)
    return snapshot_download(
        repo_id=repo_id,
        revision=revision,
        repo_type=repo_type,
        cache_dir=cache_dir,
        allow_patterns=allow_patterns,
        token=auth_token,
        **kwargs,
    )







# def get_file_from_url(
#     url: str,
#     idx: int,
#     total: int,
#     filename: Optional[str] = None,
#     output_folder: Optional[Path] = None,
#     overwrite: Optional[bool] = False,

#     # idx: int,
#     # tot: int,
#     # overwrite: bool = False,
# ):
#     output_path = output_folder / Path(url.split('/')[-1])
#     if output_path.exists() and not overwrite:
#         logger.info(f"Skipping {output_path} because it already exists.")
#         return
    
#     logger.info(f"Downloading file {idx} of {tot}...")
#     r = requests.get(
#         url, 
#         stream = True,
#         headers = {
#             'authorization': f'Bearer {huggingface_api_token}'
#         }
#     )
#     with open(output_folder / Path(url.split('/')[-1]), 'wb') as f:
#         total_size = int(r.headers.get('content-length', 0))
#         block_size = 1024
#         t = tqdm.tqdm(total=total_size, unit='iB', unit_scale=True)
#         for data in r.iter_content(block_size):
#             t.update(len(data))
#             f.write(data)
#         t.close()



from __future__ import annotations

"""
Support Lazy Loading of Modules

- Credit: https://github.com/kevdog824
"""

import importlib
from typing import Generic, TypeVar, Any
from types import ModuleType


_M = TypeVar("_M", bound=ModuleType)


class LazyLoad(Generic[_M]):
    def __init__(self, name: str, package: str | None = None) -> None:
        self._lzyname = name  # Ridiculous name avoids name clash with module
        self._lzypackage = package  # Ridiculous name avoids name clash with module
        self.__module__: ModuleType | None = None

    def __load__(self) -> _M:
        """Explicitly load the import."""
        if self.__module__ is None:
            self.__module__ = importlib.import_module(self._lzyname, self._lzypackage)
        return self.__module__

    def __reload__(self) -> _M:
        """Explicitly reload the import."""
        try:
            self.__module__ = importlib.reload(self.__module__)
        except Exception as exc:
            try:
                self.__module__ = importlib.import_module(self._lzyname, self._lzypackage)
            except Exception as e:
                raise exc from e
        return self.__module__

    def __repr__(self) -> str:
        """Gives a good representation before import is loaded.
        Uses import's __repr__ after it is loaded.
        """
        if self.__module__ is None:
            if self._lzypackage:
                return f"<Uninitialized module '{self._lzyname}' @ '{self._lzypackage}'>"
            return f"<Uninitialized module '{self._lzyname}'>"
        try:
            return self.__module__.__repr__()
        # Shouldn't happen unless someone del'd module __repr__ method for some reason
        except AttributeError:
            if self._lzypackage:
                return f"<Initialized module '{self._lzyname}' @ '{self._lzypackage}'>"
            return f"<Initialized module '{self._lzyname}'>"

    def __getattribute__(self, __name: str) -> Any:
        """Proxies attribute access to import (loads the import if not yet loaded)."""
        if __name in {
            "_lzyname",
            "_lzypackage",
            "__module__",
            "__load__",
            "__reload__",
        }:
            return super().__getattribute__(__name)
        if self.__module__ is None:
            self.__module__ = importlib.import_module(self._lzyname, self._lzypackage)
        return getattr(self.__module__, __name)


def lazy_load(name: str, package: str | None = None) -> LazyLoad:
    """Lazily load an import.
    Returns a LazyLoad object which won't load the import until you attempt
    attribute access, you call the load function with this object, or you call the
    __load__ method of this object.
    """
    return LazyLoad(name, package)


def load(__module: LazyLoad | _M) -> _M:
    """Explicitly load an import.
    If module is a LazyLoad object its wrapped import is loaded and returned.
    If module is an actual ModuleType it is simply returned.
    """
    return __module.__load__() if isinstance(__module, LazyLoad) else __module


def reload(__module: LazyLoad | _M) -> _M:
    """Explicitly reload an import.
    If module is a LazyLoad object its wrapped import is reloaded and returned.
    If module is an actual ModuleType it is reloaded and returned.
    """
    if isinstance(__module, LazyLoad):
        return __module.__reload__()
    return importlib.reload(__module)